from __future__ import print_function, division

import neo
import elephant.spike_train_generation as stg
import elephant.conversion as conv
import elephant.statistics as stat
import matplotlib.pylab as plt
import scipy.signal as ss
import time
import multiprocessing
from copy import copy
from MEArec.tools import *
import threading
import shutil
from pprint import pprint

from .simulate_cells import run_cell_model, calc_extracellular

class simulationThread(threading.Thread):
    def __init__(self, threadID, name, simulate_script, numb, tot, cell_model, model_folder, intraonly, params):
        threading.Thread.__init__(self)
        self.threadID = threadID
        self.name = name
        self.sim_script = simulate_script
        self.numb = numb
        self.tot = tot
        self.cell_model = cell_model
        self.model_folder = model_folder
        self.intra = intraonly
        self.params = params
    def run(self):
        print ("Starting " + self.name)
        print('\n\n', self.cell_model, self.numb + 1, '/', self.tot, '\n\n')
        os.system('python %s %s %s %s' \
                % (self.sim_script, join(self.model_folder, self.cell_model), self.intra, self.params))
        print ("Exiting " + self.name)


class TemplateGenerator:
    def __init__(self, *, cell_models_folder=None, templates_folder=None, temp_dict=None, info=None,
                 params=None, intraonly=False, parallel=True, delete_tmp=True):
        if temp_dict is not None and info is not None:
            self.templates = temp_dict['templates']
            self.locations = temp_dict['locations']
            self.rotations = temp_dict['rotations']
            self.celltypes = temp_dict['celltypes']
            self.info = info
        else:
            if cell_models_folder is None:
                raise AttributeError("Specify cell folder!")
            if params is None:
                print("Using default parameters")
                params = {}

            if os.path.isdir(cell_models_folder):
                cell_models = [f for f in os.listdir(join(cell_models_folder)) if 'mods' not in f]
                if len(cell_models) == 0:
                    raise AttributeError(cell_models_folder, ' contains no cell models!')
            else:
                raise NotADirectoryError('Cell models folder: does not exist!')

            this_dir, this_filename = os.path.split(__file__)
            simulate_script = join(this_dir, 'simulate_cells.py')
            params['cell_models_folder'] = cell_models_folder

            # Compile NEURON models (nrnivmodl)
            if not os.path.isdir(join(cell_models_folder, 'mods')):
                print('Compiling NEURON models')
                os.system('python %s compile %s' % (simulate_script, cell_models_folder))

            if 'sim_time' not in params.keys():
                params['sim_time'] = 1
            if 'target_spikes' not in params.keys():
                params['target_spikes'] = [3, 50]
            if 'cut_out' not in params.keys():
                params['cut_out'] = [2, 5]
            if 'dt' not in params.keys():
                params['dt'] = 2 ** -5
            if 'delay' not in params.keys():
                params['delay'] = 10
            if 'weights' not in params.keys():
                params['weights'] = [0.25, 1.75]

            if 'rot' not in params.keys():
                params['rot'] = 'physrot'
            if 'probe' not in params.keys():
                available_mea = MEA.return_mea_list()
                probe = available_mea[np.random.randint(len(available_mea))]
                print("Probe randomly set to: %s" % probe)
                params['probe'] = probe
            if 'ncontacts' not in params.keys():
                params['ncontacts'] = 1
            if 'overhang' not in params.keys():
                params['overhang'] = 1
            if 'xlim' not in params.keys():
                params['xlim'] = [10, 80]
            if 'ylim' not in params.keys():
                params['ylim'] = None
            if 'zlim' not in params.keys():
                params['zlim'] = None
            if 'offset' not in params.keys():
                params['offset'] = 0
            if 'det_thresh' not in params.keys():
                params['det_thresh'] = 30
            if 'n' not in params.keys():
                params['n'] = 50
            if 'seed' not in params.keys():
                params['seed'] = np.random.randint(1, 10000)
            elif params['seed'] is None:
                params['seed'] = np.random.randint(1, 10000)
            if templates_folder is None:
                params['templates_folder'] = os.getcwd()
                templates_folder = params['templates_folder']
            else:
                params['templates_folder'] = templates_folder

            rot = params['rot']
            n = params['n']
            probe = params['probe']
            tot = len(cell_models)

            tmp_params_path = 'tmp_params_path'
            with open(tmp_params_path, 'w') as f:
                yaml.dump(params, f)

            # Simulate neurons and EAP for different cell models sparately
            if parallel:
                start_time = time.time()
                print('Parallel')
                tot = len(cell_models)
                threads = []
                for numb, cell_model in enumerate(cell_models):
                    threads.append(simulationThread(numb, "Thread-" + str(numb), simulate_script,
                                                    numb, tot, cell_model, cell_models_folder, intraonly, tmp_params_path))
                for t in threads:
                    t.start()
                for t in threads:
                    t.join()
                print('\n\n\nSimulation time: ', time.time() - start_time, '\n\n\n')
            else:
                start_time = time.time()
                for numb, cell_model in enumerate(cell_models):
                    print('\n\n', cell_model, numb + 1, '/', len(cell_models), '\n\n')
                    os.system('python %s %s %s %s' \
                              % (simulate_script, join(cell_models_folder, cell_model), intraonly, tmp_params_path))
                print('\n\n\nSimulation time: ', time.time() - start_time, '\n\n\n')

            tmp_folder = join(templates_folder, rot, 'tmp_%d_%s' % (n, probe))
            templates, locations, rotations, celltypes = load_tmp_eap(tmp_folder)
            if delete_tmp:
                shutil.rmtree(tmp_folder)
                os.remove(tmp_params_path)

            self.info = {}

            self.templates = templates
            self.locations = locations
            self.rotations = rotations
            self.celltypes = celltypes
            self.info['params'] = params
            self.info['electrodes'] = MEA.return_mea_info(probe)


class SpikeTrainGenerator:
    def __init__(self, params=None):
        '''
        Spike Train Generator: class to create poisson or gamma spike trains

        Parameters
        ----------
        n_exc: number of excitatory cells
        n_inh: number of inhibitory cells
        f_exc: mean firing rate of excitatory cells
        f_inh: mean firing rate of inhibitory cells
        st_exc: firing rate standard deviation of excitatory cells
        st_inh: firing rate standard deviation of inhibitory cells
        process: 'poisson' - 'gamma'
        gamma_shape: shape param for gamma distribution
        t_start: starting time (s)
        t_stop: stopping time (s)
        ref_period: refractory period to remove spike violation
        n_add: number of units to add at t_add time
        t_add: time to add units
        n_remove: number of units to remove at t_remove time
        t_remove: time to remove units
        '''
        if params is None:
            print("Using default parameters")
            params = {}
        self.params = copy(params)

        if 't_start' not in self.params.keys():
            params['t_start'] = 0
        self.params['t_start'] = params['t_start'] * pq.s
        if 'duration' not in self.params.keys():
            params['duration'] = 10
        self.params['t_stop'] = self.params['t_start'] + params['duration'] * pq.s
        if 'min_rate' not in self.params.keys():
            params['min_rate'] = 0.1
        self.params['min_rate'] = params['min_rate'] * pq.Hz
        if 'ref_per' not in self.params.keys():
            params['ref_per'] = 2
        self.params['ref_per'] = params['ref_per'] * pq.ms
        if 'process' not in self.params.keys():
            params['process'] = 'poisson'
        self.params['process'] = params['process']
        if 'gamma_shape' not in self.params.keys() and params['process'] == 'gamma':
            params['gamma_shape'] = 2
            self.params['gamma_shape'] = params['gamma_shape']

        if 'rates' in self.params.keys():  # all firing rates are provided
            self.params['rates'] = self.params['rates'] * pq.Hz
            self.n_neurons = len(self.params['rates'])
        else:
            rates = []
            types = []
            if 'f_exc' not in self.params.keys():
                params['f_exc'] = 5
            self.params['f_exc'] = params['f_exc'] * pq.Hz
            if 'f_inh' not in self.params.keys():
                params['f_inh'] = 15
            self.params['f_inh'] = params['f_inh'] * pq.Hz
            if 'st_exc' not in self.params.keys():
                params['st_exc'] = 1
            self.params['st_exc'] = params['st_exc'] * pq.Hz
            if 'st_inh' not in self.params.keys():
                params['st_inh'] = 3
            self.params['st_inh'] = params['st_inh'] * pq.Hz
            if 'n_exc' not in self.params.keys():
                params['n_exc'] = 15
            self.params['n_exc'] = params['n_exc']
            if 'n_inh' not in self.params.keys():
                params['n_inh'] = 5
            self.params['n_inh'] = params['n_inh']

            for exc in range(self.params['n_exc']):
                rate = self.params['st_exc'] * np.random.randn() + self.params['f_exc']
                if rate < self.params['min_rate']:
                    rate = self.params['min_rate']
                rates.append(rate)
                types.append('e')
            for inh in range(self.params['n_inh']):
                rate = self.params['st_inh'] * np.random.randn() + self.params['f_inh']
                if rate < self.params['min_rate']:
                    rate = self.params['min_rate']
                rates.append(rate)
                types.append('i')
            self.params['rates'] = rates
            self.params['types'] = types
            self.n_neurons = len(self.params['rates'])

        np.random.seed(self.params['seed'])

        self.changing = False
        self.intermittent = False

        self.info = params

        #TODO add spike synchrony and overlapping synchrony
        #self.sync_rate = float(sync)
        #
        # self.save = save
        #
        # #print( 'Modulation: ', self.modulation
        #

        # self.changing = False
        # self.n_add = n_add
        # self.n_remove = n_remove
        # self.t_add = int(t_add) * pq.s
        # self.t_remove = int(t_remove) * pq.s
        #
        # self.intermittent = False
        # self.n_int = n_int
        # self.t_int = t_int
        # self.t_burst = t_burst
        # self.t_int_sd = t_int
        # self.t_burst_sd = t_burst
        # self.f_int = f_int
        #
        # self.idx = 0
        #
        #
        # if n_add != 0:
        #     if t_add == 0:
        #         raise Exception('Provide time to add units')
        #     else:
        #         self.changing = True
        # if n_remove != 0:
        #     if t_remove == 0:
        #         raise Exception('Provide time to remove units')
        #     else:
        #         self.changing = True
        # if n_int != 0:
        #     self.intermittent = True
        #
        #
        # if self.changing:
        #     n_tot = n_exc + n_inh
        #     perm_idxs = np.random.permutation(np.arange(n_tot))
        #     self.idxs_add = perm_idxs[:self.n_add]
        #     self.idxs_remove = perm_idxs[-self.n_remove:]
        # else:
        #     self.idxs_add = []
        #     self.idxs_remove = []
        #
        # if self.intermittent:
        #     n_tot = n_exc + n_inh
        #     perm_idxs = np.random.permutation(np.arange(n_tot))
        #     self.idxs_int = perm_idxs[:self.n_int]
        # else:
        #     self.idxs_int = []

    def set_spiketrain(self, idx, spiketrain):
        '''
        Sets spike train idx to new spiketrain
        Parameters
        ----------
        idx: index of spike train to set
        spiketrain: new spike train

        Returns
        -------

        '''
        self.all_spiketrains[idx] = spiketrain

    def generate_spikes(self):
        '''
        Generate spike trains based on default_params of the SpikeTrainGenerator class.
        self.all_spiketrains contains the newly generated spike trains

        Returns
        -------

        '''

        self.all_spiketrains = []
        idx = 0
        for n in range(self.n_neurons):
            if not self.changing and not self.intermittent:
                rate = self.params['rates'][n]
                if self.params['process'] == 'poisson':
                    st = stg.homogeneous_poisson_process(rate,
                                                         self.params['t_start'], self.params['t_stop'])
                elif self.params['process'] == 'gamma':
                    st = stg.homogeneous_gamma_process(self.params['gamma_shape'], rate, self.params['rates'][n],
                                                       self.params['t_start'], self.params['t_stop'])
            else:
                raise NotImplementedError('Changing and intermittent spiketrains are not impleented yet')
            self.all_spiketrains.append(st)
            self.all_spiketrains[-1].annotate(freq=rate)
            if 'n_exc' in self.params.keys() and 'n_inh' in self.params.keys():
                if idx < self.params['n_exc']:
                    self.all_spiketrains[-1].annotate(type='E')
                else:
                    self.all_spiketrains[-1].annotate(type='I')
            idx += 1

        # check consistency and remove spikes below refractory period
        for idx, st in enumerate(self.all_spiketrains):
            isi = stat.isi(st)
            idx_remove = np.where(isi < self.params['ref_per'])[0]
            spikes_to_remove = len(idx_remove)
            unit = st.times.units

            while spikes_to_remove > 0:
                new_times = np.delete(st.times, idx_remove[0]) * unit
                st = neo.SpikeTrain(new_times, t_start=self.params['t_start'], t_stop=self.params['t_stop'])
                isi = stat.isi(st)
                idx_remove = np.where(isi < self.params['ref_per'])[0]
                spikes_to_remove = len(idx_remove)

            st.annotations = self.all_spiketrains[idx].annotations
            self.set_spiketrain(idx, st)

    def raster_plots(self, marker='|', markersize=5, mew=2):
        '''
        Plots raster plots of spike trains

        Parameters
        ----------
        marker: marker type (def='|')
        markersize: marker size (def=5)
        mew: marker edge width (def=2)

        Returns
        -------
        ax: matplotlib axes

        '''
        fig = plt.figure()
        ax = fig.add_subplot(111)
        for i, spiketrain in enumerate(self.all_spiketrains):
            t = spiketrain.rescale(pq.s)
            if i < self.params['n_exc']:
                ax.plot(t, i * np.ones_like(t), color='b', marker=marker, ls='', markersize=markersize, mew=mew)
            else:
                ax.plot(t, i * np.ones_like(t), color='r', marker=marker, ls='', markersize=markersize, mew=mew)
        ax.axis('tight')
        ax.set_xlim([self.params['t_start'].rescale(pq.s), self.params['t_stop'].rescale(pq.s)])
        ax.set_xlabel('Time (ms)', fontsize=16)
        ax.set_ylabel('Spike Train Index', fontsize=16)
        plt.gca().tick_params(axis='both', which='major', labelsize=14)

        return ax

    def resample_spiketrains(self, fs=None, T=None):
        '''
        Resamples spike trains. Provide either fs or T parameters
        Parameters
        ----------
        fs: new sampling frequency (quantity)
        T: new period (quantity)

        Returns
        -------
        matrix with resampled binned spike trains

        '''
        resampled_mat = []
        if not fs and not T:
            print('Provide either sampling frequency fs or time period T')
        elif fs:
            if not isinstance(fs, Quantity):
                raise ValueError("fs must be of type pq.Quantity")
            binsize = 1. / fs
            binsize.rescale('ms')
            resampled_mat = []
            for sts in self.all_spiketrains:
                spikes = conv.BinnedSpikeTrain(sts, binsize=binsize).to_array()
                resampled_mat.append(np.squeeze(spikes))
        elif T:
            binsize = T
            if not isinstance(T, Quantity):
                raise ValueError("T must be of type pq.Quantity")
            binsize.rescale('ms')
            resampled_mat = []
            for sts in self.all_spiketrains:
                spikes = conv.BinnedSpikeTrain(sts, binsize=binsize).to_array()
                resampled_mat.append(np.squeeze(spikes))
        return np.array(resampled_mat)

    def add_synchrony(self, idxs, rate=0.05):
        '''
        Adds synchronous spikes between pairs of spike trains at a certain rate
        Parameters
        ----------
        idxs: list or array with the 2 indices
        rate: probability of adding a synchronous spike to spike train idxs[1] for each spike of idxs[0]

        Returns
        -------

        '''
        idx1 = idxs[0]
        idx2 = idxs[1]
        st1 = self.all_spiketrains[idx1]
        st2 = self.all_spiketrains[idx2]
        times2 = st2.times
        t_start = st2.t_start
        t_stop = st2.t_stop
        unit = times2.units
        for t1 in st1:
            rand = np.random.rand()
            if rand <= rate:
                # check time difference
                t_diff = np.abs(t1.rescale(pq.ms).magnitude - times2.rescale(pq.ms).magnitude)
                if np.all(t_diff > self.params['ref_period']):
                    times2 = np.sort(np.concatenate((np.array(times2), np.array([t1]))))
                    times2 = times2 * unit
                    st2 = neo.SpikeTrain(times2, t_start=t_start, t_stop=t_stop)
                    self.set_spiketrain(idx2, st2)

    def bursting_st(self, freq=None, min_burst=3, max_burst=10):
        pass

class RecordingGenerator:
    def __init__(self, spgen=None, tempgen=None, params=None, rec_dict=None, info=None):
        '''

        Parameters
        ----------
        spgen
        tempgen
        params
        rec_dict
        info
        '''
        if rec_dict is not None and info is not None:
            self.recordings = rec_dict['recordings']
            self.spiketrains = rec_dict['spiketrains']
            self.templates = rec_dict['templates']
            self.positions = rec_dict['positions']
            self.times = rec_dict['times']
            self.peaks = rec_dict['peaks']
            self.sources = rec_dict['sources']
            self.info = info
        else:
            if spgen is None or tempgen is None:
                raise AttributeError("Specify SpikeGenerator and TemplateGenerator objects!")
            if params is None:
                print("Using default parameters")
                params = {'spiketrains': {},
                          'celltypes': {},
                          'templates': {},
                          'recordings': {}}
            self.params = copy(params)

            temp_params = self.params['templates']
            rec_params = self.params['recordings']
            st_params = self.params['spiketrains']
            celltype_params = self.params['cell_types']

            eaps = tempgen.templates
            locs = tempgen.locations
            rots = tempgen.rotations
            celltypes = tempgen.celltypes
            temp_info = tempgen.info

            spiketrains = spgen.all_spiketrains

            n_neurons = len(spiketrains)
            cut_outs = temp_info['params']['cut_out']
            duration = spiketrains[0].t_stop - spiketrains[0].t_start

            if 'fs' not in rec_params.keys():
                params['recordings']['fs'] = 1. / temp_info['params']['dt']
            elif params['recordings']['fs'] is None:
                params['recordings']['fs'] = 1. / temp_info['params']['dt']
            fs = params['recordings']['fs'] * pq.kHz

            if 'noise_level' not in rec_params.keys():
                params['recordings']['noise_level'] = 15
            noise_level =  params['recordings']['noise_level']

            if 'noise_mode' not in rec_params.keys():
                params['recordings']['noise_mode'] = 'uncorrelated'
            noise_mode = params['recordings']['noise_mode']

            if 'filter' not in rec_params.keys():
                params['recordings']['filter'] = True
            filter = params['recordings']['noise_mode']

            if 'cutoff' not in rec_params.keys():
                params['recordings']['cutoff'] = [300., 6000.]
            cutoff = params['recordings']['cutoff'] * pq.Hz

            if 'modulation' not in rec_params.keys():
                params['recordings']['modulation'] = 'electrode'
            modulation = params['recordings']['modulation']

            if 'chunk_duration' not in rec_params.keys():
                params['recordings']['chunk_duration'] = 0
            chunk_duration = params['recordings']['chunk_duration'] * pq.s

            if 'mrand' not in rec_params.keys():
                params['recordings']['mrand'] = 1
            mrand = params['recordings']['mrand']

            if 'sdrand' not in rec_params.keys():
                params['recordings']['sdrand'] = 0.05
            sdrand = params['recordings']['sdrand']

            if 'overlap' not in rec_params.keys():
                params['recordings']['overlap'] = False
            overlap = params['recordings']['overlap']

            if 'seed' not in rec_params.keys():
                params['recordings']['seed'] = np.random.randint(1,1000)
            elif params['recordings']['seed'] is None:
                params['recordings']['seed'] = np.random.randint(1,1000)
            noise_seed = params['recordings']['seed']

            if 'x_lim' not in temp_params.keys():
                params['templates']['x_lim'] = None
            x_lim =  params['templates']['x_lim']

            if 'y_lim' not in temp_params.keys():
                params['templates']['y_lim'] = None
            y_lim = params['templates']['y_lim']

            if 'z_lim' not in temp_params.keys():
                params['templates']['z_lim'] = None
            z_lim = params['templates']['z_lim']

            if 'min_amp' not in temp_params.keys():
                params['templates']['min_amp'] = 50
            min_amp = params['templates']['min_amp']

            if 'min_dist' not in temp_params.keys():
                params['templates']['min_dist'] = 25
            min_dist = params['templates']['min_dist']

            if 'overlap_threshold' not in temp_params.keys():
                params['templates']['overlap_threshold'] = 0.8
            overlap_threshold = params['templates']['overlap_threshold']

            if 'pad_len' not in temp_params.keys():
                params['templates']['pad_len'] = [3., 3.]
            pad_len = params['templates']['pad_len']

            if 'n_jitters' not in temp_params.keys():
                params['templates']['n_jitters'] = 10
            n_jitters = params['templates']['n_jitters']

            if 'upsample' not in temp_params.keys():
                params['templates']['upsample'] = 8
            upsample = params['templates']['upsample']

            if 'seed' not in temp_params.keys():
                params['templates']['seed'] = np.random.randint(1,1000)
            elif params['templates']['seed'] is None:
                params['templates']['seed'] = np.random.randint(1,1000)
            temp_seed = params['templates']['seed']

            parallel=False

            if 'excitatory' in celltype_params.keys() and 'inhibitory' in celltype_params.keys():
                exc_categories = celltype_params['excitatory']
                inh_categories = celltype_params['inhibitory']
                bin_cat = get_binary_cat(celltypes, exc_categories, inh_categories)
            else:
                bin_cat = np.array(['U'] * len(celltypes))

            # load MEA info
            electrode_name = temp_info['electrodes']['electrode_name']
            elinfo = MEA.return_mea_info(electrode_name)
            offset = temp_info['params']['offset']
            elinfo['offset'] = offset
            mea = MEA.return_mea(info=elinfo)
            mea_pos = mea.positions

            params['recordings'].update({'electrode_name': electrode_name,
                                         'duration': float(duration.magnitude),
                                         'fs': float(fs.rescale('Hz').magnitude),
                                         'n_neurons': n_neurons})


            n_elec = eaps.shape[1]
            # this is fixed from recordings
            spike_duration = np.sum(temp_info['params']['cut_out'])*pq.ms
            spike_fs = 1./temp_info['params']['dt']*pq.kHz

            print('Selecting cells')
            if 'type' in spiketrains[0].annotations.keys():
                n_exc = [st.annotations['type'] for st in spiketrains].count('E')
                n_inh = n_neurons - n_exc

            np.random.seed(temp_seed)
            idxs_cells = select_templates(locs, eaps, bin_cat, n_exc, n_inh, x_lim=x_lim, y_lim=y_lim, z_lim=z_lim,
                                          min_amp=min_amp, min_dist=min_dist, verbose=False)
            template_celltypes = celltypes[idxs_cells]
            template_locs = locs[idxs_cells]
            templates_bin = bin_cat[idxs_cells]
            templates = eaps[idxs_cells]

            # peak images
            peak = []
            for tem in templates:
                dt = 2 ** -5
                feat = get_EAP_features(tem, ['Na'], dt=dt)
                peak.append(-np.squeeze(feat['na']))
            peak = np.array(peak)

            up = fs
            down = spike_fs
            sampling_ratio = float(up/down)
            # resample spikes
            resample=False
            pad_samples = [int((pp*fs).magnitude) for pp in pad_len]
            n_resample = int((fs * spike_duration).magnitude)
            if templates.shape[2] != n_resample:
                templates_pol = np.zeros((templates.shape[0], templates.shape[1], n_resample))
                print('Resampling spikes')
                for t, tem in enumerate(templates):
                    tem_pad = np.pad(tem, [(0,0), pad_samples], 'edge')
                    tem_poly = ss.resample_poly(tem_pad, up, down, axis=1)
                    templates_pol[t, :] = tem_poly[:, int(sampling_ratio*pad_samples[0]):int(sampling_ratio*pad_samples[0])
                                                                                         +n_resample]
                resample=True
            else:
                templates_pol = templates

            templates_pad = []
            templates_spl = []
            print('Padding template edges')
            for t, tem in enumerate(templates_pol):
                tem, _ = cubic_padding(tem, pad_len, fs)
                templates_pad.append(tem)

            print('Creating time jittering')
            jitter = 1. / fs
            templates_jitter = []
            for temp in templates_pad:
                temp_up = ss.resample_poly(temp, upsample, 1., axis=1)
                nsamples_up = temp_up.shape[1]
                temp_jitt = []
                for n in range(n_jitters):
                    # align waveform
                    shift = int((jitter * (np.random.random() - 0.5) * upsample * fs).magnitude)
                    if shift > 0:
                        t_jitt = np.pad(temp_up, [(0, 0), (np.abs(shift), 0)], 'constant')[:, :nsamples_up]
                    elif shift < 0:
                        t_jitt = np.pad(temp_up, [(0, 0), (0, np.abs(shift))], 'constant')[:, -nsamples_up:]
                    else:
                        t_jitt = temp_up
                    temp_down = ss.decimate(t_jitt, upsample, axis=1)
                    temp_jitt.append(temp_down)
                templates_jitter.append(temp_jitt)
            templates = np.array(templates_jitter)

            cut_outs_samples = np.array(cut_outs * fs.rescale('kHz').magnitude, dtype=int) + pad_samples

        #
            # if self.sync_rate != 0:
            #     #print( 'Adding synchrony on overlapping spikes'
            #     self.overlapping = find_overlapping_spikes(self.templates, thresh=self.overlap_threshold)
            #
            #     for over in self.overlapping:
            #         self.spgen.add_synchrony(over, rate=self.sync_rate)
            # else:
            #     self.overlapping = []

            # find SNR and annotate
            print('Computing spike train SNR')
            for t_i, temp in enumerate(templates):
                min_peak = np.min(temp)
                snr = np.abs(min_peak/float(noise_level))
                spiketrains[t_i].annotate(snr=snr)
                # print( min_peak, snr)
        #
        #
        #     if self.plot_figures and self.sync_rate != 0:
        #         ax = self.spgen.raster_plots()
        #         ax.set_title('After synchrony')
        #
            print('Adding spiketrain annotations')
            for i, st in enumerate(spiketrains):
                st.annotate(bintype=templates_bin[i], mtype=template_celltypes[i], loc=template_locs[i])
            if overlap:
                print('Finding temporally overlapping spikes')
                overlapping = find_overlapping_templates(templates, thresh=overlap_threshold)
                annotate_overlapping_spikes(spiketrains, overlapping_pairs=overlapping, verbose=True)


            amp_mod = []
            cons_spikes = []
            exp = 0.3
            n_isi = 5
            mem_isi = 10*pq.ms
            if modulation == 'isi':
                print('ISI modulation')
                for st in spiketrains:
                    amp, cons = ISI_amplitude_modulation(st, mrand=mrand, sdrand=sdrand,
                                                         n_spikes=n_isi, exp=exp, mem_ISI=mem_isi)
                    amp_mod.append(amp)
                    cons_spikes.append(cons)
            elif modulation == 'template':
                print('Template modulation')
                for st in spiketrains:
                    amp, cons = ISI_amplitude_modulation(st, mrand=mrand, sdrand=sdrand,
                                                         n_spikes=0, exp=exp, mem_ISI=mem_isi)
                    amp_mod.append(amp)
                    cons_spikes.append(cons)
            elif modulation == 'electrode':
                print('Electrode modulaton')
                for st in spiketrains:
                    amp, cons = ISI_amplitude_modulation(st, n_el=n_elec, mrand=mrand, sdrand=sdrand,
                                                         n_spikes=0, exp=exp, mem_ISI=mem_isi)
                    amp_mod.append(amp)
                    cons_spikes.append(cons)

            spike_matrix = resample_spiketrains(spiketrains, fs=fs)
            n_samples = spike_matrix.shape[1]

            #print( 'Generating clean recordings'
            recordings = np.zeros((n_elec, n_samples))
            times = np.arange(recordings.shape[1]) / fs


            # modulated convolution
            pool = multiprocessing.Pool(n_neurons)
            t_start = time.time()
            gt_spikes = []

            # divide in chunks
            chunks = []
            if duration > chunk_duration and chunk_duration != 0:
                start = 0*pq.s
                finished = False
                while not finished:
                    chunks.append([start, start+chunk_duration])
                    start=start+chunk_duration
                    if start >= duration:
                        finished = True
                print('Chunks: ', chunks)


            if len(chunks) > 0:
                recording_chunks = []
                for ch, chunk in enumerate(chunks):
                    #print( 'Generating chunk ', ch+1, ' of ', len(chunks)
                    idxs = np.where((times>=chunk[0]) & (times<chunk[1]))[0]
                    spike_matrix_chunk = spike_matrix[:, idxs]
                    rec_chunk=np.zeros((n_elec, len(idxs)))
                    amp_chunk = []
                    for i, st in enumerate(spiketrains):
                        idxs = np.where((st >= chunk[0]) & (st < chunk[1]))[0]
                        if modulation != 'none':
                            amp_chunk.append(amp_mod[i][idxs])

                    if not parallel:
                        for st, spike_bin in enumerate(spike_matrix_chunk):
                            #print( 'Convolving with spike ', st, ' out of ', spike_matrix_chunk.shape[0]
                            if modulation == 'none':
                                rec_chunk += convolve_templates_spiketrains(st, spike_bin, templates[st],
                                                                            cut_out=cut_outs_samples)
                            else:
                                rec_chunk += convolve_templates_spiketrains(st, spike_bin, templates[st],
                                                                            cut_out=cut_outs_samples,
                                                                                       modulation=True,
                                                                                       amp_mod=amp_chunk[st])
                    else:
                        if modulation == 'none':
                            results = [pool.apply_async(convolve_templates_spiketrains, (st, spike_bin, templates[st],))
                                       for st, spike_bin in enumerate(spike_matrix_chunk)]
                        else:
                            results = [pool.apply_async(convolve_templates_spiketrains,
                                                        (st, spike_bin, templates[st], True, amp))
                                       for st, (spike_bin, amp) in enumerate(zip(spike_matrix_chunk, amp_chunk))]
                        for r in results:
                            rec_chunk += r.get()

                    recording_chunks.append(rec_chunk)
                recordings = np.hstack(recording_chunks)
            else:
                for st, spike_bin in enumerate(spike_matrix):
                    print('Convolving with spike ', st, ' out of ', spike_matrix.shape[0])
                    if modulation == 'none':
                        # reset random seed to keep sampling of jitter spike same
                        seed = np.random.randint(10000)
                        np.random.seed(seed)

                        recordings += convolve_templates_spiketrains(st, spike_bin, templates[st],
                                                                     cut_out=cut_outs_samples)
                        np.random.seed(seed)
                        gt_spikes.append(convolve_single_template(st, spike_bin,
                                                                  templates[st, :, np.argmax(peak[st])],
                                                                  cut_out=cut_outs_samples))
                    elif modulation == 'electrode':
                        seed = np.random.randint(10000)
                        np.random.seed(seed)
                        recordings += convolve_templates_spiketrains(st, spike_bin, templates[st], cut_out=cut_outs_samples,
                                                                          modulation=True,
                                                                          amp_mod=amp_mod[st])
                        np.random.seed(seed)
                        gt_spikes.append(convolve_single_template(st, spike_bin,
                                                                  templates[st, :, np.argmax(peak[st])],
                                                                  cut_out=cut_outs_samples,
                                                                  modulation=True,
                                                                  amp_mod=amp_mod[st][:,
                                                                          np.argmax(peak[st])]))
                    elif modulation == 'template' or modulation == 'all':
                        seed = np.random.randint(10000)
                        np.random.seed(seed)
                        recordings += convolve_templates_spiketrains(st, spike_bin, templates[st], cut_out=cut_outs_samples,
                                                                          modulation=True,
                                                                          amp_mod=amp_mod[st])
                        np.random.seed(seed)
                        gt_spikes.append(convolve_single_template(st, spike_bin,
                                                                  templates[st, :, np.argmax(peak[st])],
                                                                  cut_out=cut_outs_samples,
                                                                  modulation=True,
                                                                  amp_mod=amp_mod[st]))

            pool.close()
            gt_spikes = np.array(gt_spikes)

            print('Elapsed time ', time.time() - t_start)
            clean_recordings = copy(recordings)

            print('Adding noise')
            np.random.seed(noise_seed)
            if noise_level > 0:
                if noise_mode == 'uncorrelated':
                    additive_noise = noise_level * np.random.randn(recordings.shape[0],
                                                                             recordings.shape[1])
                    recordings += additive_noise
                elif noise_mode == 'correlated-dist':
                    # TODO divide in chunks
                    cov_dist = np.zeros((n_elec, n_elec))
                    for i, el in enumerate(mea_pos):
                        for j, p in enumerate(mea_pos):
                            if i != j:
                                cov_dist[i, j] = (0.5*np.min(mea_pitch))/np.linalg.norm(el - p)
                            else:
                                cov_dist[i, j] = 1

                    additive_noise = np.random.multivariate_normal(np.zeros(n_elec), cov_dist,
                                                                   size=(recordings.shape[0], recordings.shape[1]))
                    recordings += additive_noise
                elif noise_level == 'experimental':
                    pass
                    #print( 'experimental noise model'
            else:
                print('Noise level is set to 0')

            if filter:
                print('Filtering signals')
                if fs/2. < cutoff[1]:
                    recordings = filter_analog_signals(recordings, freq=cutoff[0], fs=fs, filter_type='highpass')
                else:
                    recordings = filter_analog_signals(recordings, freq=cutoff, fs=fs)

            print('Extracting spike waveforms')
            extract_wf(spiketrains, recordings, times, fs)

            self.recordings = recordings
            self.times = times
            self.positions = mea_pos
            self.templates = templates
            self.spiketrains = spiketrains
            self.peaks = peak
            self.sources = gt_spikes
            self.info = params

def gen_recordings(params=None, templates=None, tempgen=None):
    '''

    Parameters
    ----------
    templates: str
        Path to generated templates
    params: dict OR str
        Dictionary containing recording parameters OR path to yaml file containing parameters

    Returns
    -------
    recgen: RecordingGenerator

    '''
    if isinstance(params, str):
        if os.path.isfile(params) and (params.endswith('yaml') or params.endswith('yml')):
            with open(params, 'r') as pf:
                params_dict = yaml.load(pf)
    elif isinstance(params, dict):
        params_dict = params
    else:
        params_dict = {}

    if 'spiketrains' not in params_dict:
        params_dict['spiketrains'] = {}
    if 'templates' not in params_dict:
        params_dict['templates'] = {}
    if 'recordings' not in params_dict:
        params_dict['recordings'] = {}
    if 'cell_types' not in params_dict:
        params_dict['cell_types'] = {}


    if tempgen is None and templates is None:
        raise AttributeError("Provide either 'templates' or 'tempgen' TemplateGenerator object")

    if tempgen is None:
        if os.path.isdir(templates):
            tempgen = load_templates(templates, verbose=False)
        elif templates.endswith('h5') or templates.endswith('hdf5'):
            tempgen = load_templates(templates, verbose=False)
        else:
            raise AttributeError("'templates' is not a folder or an hdf5 file")

    if 'seed' in params_dict['spiketrains']:
        if params_dict['spiketrains']['seed'] is None:
            params_dict['spiketrains']['seed'] = np.random.randint(1, 10000)
    else:
        params_dict['spiketrains'].update({'seed': np.random.randint(1, 10000)})

    if 'seed' in params_dict['templates']:
        if params_dict['templates']['seed'] is None:
            params_dict['templates']['seed'] = np.random.randint(1, 10000)
    else:
        params_dict['templates'].update({'seed': np.random.randint(1, 10000)})

    if 'seed' in params_dict['recordings']:
        if params_dict['recordings']['seed'] is None:
            params_dict['recordings']['seed'] = np.random.randint(1, 10000)
    else:
        params_dict['recordings'].update({'recordings': np.random.randint(1, 10000)})

    # Generate spike trains
    spgen = SpikeTrainGenerator(params_dict['spiketrains'])
    spgen.generate_spikes()
    spiketrains = spgen.all_spiketrains

    params_dict['spiketrains'] = spgen.info
    # Generate recordings
    recgen = RecordingGenerator(spgen, tempgen, params_dict)

    return recgen


def gen_templates(cell_models_folder, params=None, templates_folder=None,
                  intraonly=False, parallel=True, delete_tmp=True):
    '''

    Parameters
    ----------
    tmp_params_path: str
        path to tmp params.yaml (same as templates_params with extra 'simulate_script', 'templates_folder' fields)
    intraonly: bool
        if True only intracellular simulation is run
    parallel:
        if True multi-threading is used
    delete_tmp:
        if True the temporary files are deleted

    Returns
    -------
    templates: np.array
        Array containing num_eaps x num_channels templates
    locations: np.array
        Array containing num_eaps x 3 (x,y,z) soma positions
    rotations: np.array
        Array containing num_eaps x 3 (x_rot, y_rot. z_rot) rotations
    celltypes: np.array
        Array containing num_eaps cell-types

    '''
    if isinstance(params, str):
        if os.path.isfile(params) and (params.endswith('yaml') or params.endswith('yml')):
            with open(params, 'r') as pf:
                params_dict = yaml.load(pf)
    elif isinstance(params, dict):
        params_dict = params
    else:
        params_dict=None

    if templates_folder is not None:
        if not os.path.isdir(templates_folder):
            os.makedirs(templates_folder)

    tempgen = TemplateGenerator(cell_models_folder=cell_models_folder,
                                params=params_dict,
                                templates_folder=templates_folder,
                                intraonly=intraonly,
                                parallel=parallel,
                                delete_tmp=delete_tmp)

    return tempgen


from MEArec.tools import load_templates, load_recordings, hdf5_to_templates, \
    hdf5_to_recordings, templates_to_hdf5, recordings_to_hdf5, save_recording_generator, save_template_generator
from MEArec.generators import gen_recordings, gen_templates, RecordingGenerator, TemplateGenerator


__version__ = '1.0.2'


import os
from os.path import join
import time
import MEAutility as MEA
import click
import numpy as np
import yaml
import shutil
import MEArec.generators as generators
from MEArec import recordings_to_hdf5, templates_to_hdf5, hdf5_to_recordings, hdf5_to_templates
import pprint
import time


def getDefaultConfig():
    this_dir, this_filename = os.path.split(__file__)
    home = os.path.expanduser("~")
    mearec_home = join(home, '.config', 'mearec')
    if not os.path.isdir(mearec_home):
        os.makedirs(mearec_home)
        shutil.copytree(join(this_dir, 'default_params'), join(mearec_home, 'default_params'))
        default_info = {'templates_params': join(mearec_home, 'default_params', 'templates_params.yaml'),
                        'recordings_params': join(mearec_home, 'default_params', 'recordings_params.yaml'),
                        'templates_folder': join(mearec_home, 'templates'),
                        'recordings_folder': join(mearec_home, 'recordings'),
                        'cell_models_folder': ''}
        with open(join(mearec_home, 'mearec.conf'), 'w') as f:
            yaml.dump(default_info, f)
    else:
        with open(join(mearec_home, 'mearec.conf'), 'r') as f:
            default_info = yaml.load(f)
    return default_info, mearec_home

@click.group()
def cli():
    """MEArec: Fast and customizable simulation of extracellular recordings on Multi-Electrode-Arrays """
    pass


@cli.command()
@click.option('--params', '-prm', default=None,
              help='path to default_params.yaml (otherwise default default_params are used and some of the parameters'
                   'can be overwritten with the following options)')
@click.option('--default', is_flag=True,
              help='shows default values for simulation')
@click.option('--fname', '-fn', default=None,
              help='template filename')
@click.option('--folder', '-fol', default=None,
              help='templates output base folder')
@click.option('--cellfolder', '-cf', default=None,
              help='folder containing bbp cell models')
@click.option('--rot', '-r', default=None,
              help='possible rotation arguments: Norot-physrot-3drot (default=physrot)')
@click.option('--probe', '-p', default=None,
              help='probe name from available electrodes (default=None)')
@click.option('--intraonly', '-i', default=False, type=bool, is_flag=True,
              help='if True it only simulate intracellular (default=False)')
@click.option('--n', '-n', default=None, type=int,
              help='number of observations per cell type (default=1000)')
@click.option('--ncontacts', '-nc', default=None, type=int,
              help='number of contacts per electrode (default=1)')
@click.option('--overhang', '-ov', default=None, type=float,
              help='extension (um) beyond MEA boundaries for neuron locations (default=30.)')
@click.option('--offset', '-off', default=None, type=float,
              help='x_plane (um) coordinate for MEA (default=0)')
@click.option('--xlim', '-xl', default=None,  nargs=2, type=float,
              help='limits ( low high ) for neuron locations in the x-axis (depth) (default=[10.,80.])')
@click.option('--ylim', '-yl', default=None,  nargs=2, type=float,
              help='limits ( low high ) for neuron locations in the y-axis (default=None)')
@click.option('--xlim', '-xl', default=None,  nargs=2, type=float,
              help='limits ( low high ) for neuron locations in the z-axis (default=None)')
@click.option('--det-thresh', '-dt', default=None, type=float,
              help='detection threshold for EAPs (default=30)')
@click.option('--seed', '-s', default=None, type=int,
              help='random seed for template generation (int)')
@click.option('--intraonly', '-io', is_flag=True,
              help='only run intracellular simulations')
@click.option('--parallel', '-par', is_flag=True,
              help='run with multiprocessing tool')
def gen_templates(params, **kwargs):
    """Generates EAP templates on multi-electrode arrays using biophyical NEURON simulations and LFPy"""
    this_dir, this_filename = os.path.split(__file__)
    info, config_folder = getDefaultConfig()

    if params is None:
        with open(info['templates_params'], 'r') as pf:
            params_dict = yaml.load(pf)
    else:
        with open(params, 'r') as pf:
            params_dict = yaml.load(pf)

    if kwargs['default'] is True:
        pprint.pprint(params_dict)
        MEA.return_mea()
        return

    if kwargs['cellfolder'] is not None:
        model_folder = kwargs['cellfolder']
    else:
        model_folder = info['cell_models_folder']
    params_dict['cell_models_folder'] = model_folder

    if kwargs['seed'] is not None:
        seed = kwargs['seed']
    else:
        seed = np.random.randint(1, 10000)
    params_dict['seed'] = seed


    if kwargs['folder'] is not None:
        params_dict['templates_folder'] = kwargs['folder']
    else:
        templates_folder = info['templates_folder']
    intraonly = kwargs['intraonly']

    if kwargs['rot'] is not None:
        params_dict['rot'] = kwargs['rot']
    if kwargs['n'] is not None:
        params_dict['n'] = kwargs['n']
    if kwargs['ncontacts'] is not None:
        params_dict['ncontacts'] = kwargs['ncontacts']
    if kwargs['overhang'] is not None:
        params_dict['overhang'] = kwargs['overhang']
    if kwargs['offset'] is not None:
        params_dict['offset'] = kwargs['offset']
    if kwargs['det_thresh'] is not None:
        params_dict['det_thresh'] = kwargs['det_thresh']
    if kwargs['probe'] is not None:
        params_dict['probe'] = kwargs['probe']
    else:
        intraonly = True
    parallel = kwargs['parallel']
    params_dict['templates_folder'] = templates_folder

    tempgen = generators.gen_templates(model_folder, params_dict, templates_folder, intraonly, parallel)

    # Merge simulated data and cleanup
    if not intraonly:
        rot = params_dict['rot']
        n = params_dict['n']
        probe = params_dict['probe']
        if kwargs['fname'] is None:
            fname = 'templates_%d_%s_%s' % (n, probe, time.strftime("%d-%m-%Y"))
        else:
            fname = kwargs['fname']
        save_folder = join(templates_folder, rot, fname)
        if not os.path.isdir(save_folder):
            os.makedirs(save_folder)
        np.save(join(save_folder, 'templates'), tempgen.templates)
        np.save(join(save_folder, 'locations'), tempgen.locations)
        np.save(join(save_folder, 'rotations'), tempgen.rotations)
        np.save(join(save_folder, 'celltypes'), tempgen.celltypes)
        info = tempgen.info
        yaml.dump(info, open(join(save_folder, 'info.yaml'), 'w'), default_flow_style=False)
        print('\nSaved eap templates in', save_folder, '\n')



@cli.command()
@click.option('--templates', '-t', default=None,
              help='eap templates path')
@click.option('--params', '-prm', default=None,
              help='path to default_params.yaml (otherwise default default_params are used and some of the parameters can be overwritten with the following options)')
@click.option('--default', is_flag=True,
              help='shows default values for simulation')
@click.option('--fname', '-fn', default=None,
              help='recording filename')
@click.option('--folder', '-fol', default=None,
              help='recording output base folder')
@click.option('--duration', '-d', default=None, type=float,
              help='duration in s (default=10)')
@click.option('--n-exc', '-ne', default=None, type=int,
              help='number of excitatory cells (default=15)')
@click.option('--n-inh', '-ni', default=None, type=int,
              help='number of inhibitory cells (default=5)')
@click.option('--f-exc', '-fe', default=None, type=float,
              help='average firing rate of excitatory cells in Hz (default=5)')
@click.option('--f-inh', '-fi', default=None, type=float,
              help='average firing rate of inhibitory cells in Hz (default=15)')
@click.option('--st-exc', '-se', default=None, type=float,
              help='firing rate standard deviation of excitatory cells in Hz (default=1)')
@click.option('--st-inh', '-si', default=None, type=float,
              help='firing rate standard deviation of inhibitory cells in Hz (default=3)')
@click.option('--min-rate', '-mr', default=None, type=float,
              help='minimum firing rate (default=0.5)')
@click.option('--ref-per', '-rp', default=None, type=float,
              help='refractory period in ms (default=2)')
@click.option('--process', '-p', default='poisson', type=click.Choice(['poisson', 'gamma']),
              help='process for generating spike trains (default=poisson)')
@click.option('--tstart', default=None, type=float,
              help='start time in s (default=0)')
@click.option('--min-dist', '-md', default=None, type=int,
              help='minumum distance between neuron in um (default=25)')
@click.option('--min-amp', '-ma', default=None, type=int,
              help='minumum eap amplitude in uV (default=50)')
@click.option('--fs', default=None, type=float,
              help='sampling frequency in kHz (default from templates sampling frequency)')
@click.option('--noise-lev', '-nl', default=None, type=int,
              help='noise level in uV (default=10)')
@click.option('--modulation', '-m', default=None, type=click.Choice(['none', 'template', 'electrode']),
              help='modulation type')
@click.option('--chunk', '-ch', default=None, type=float,
              help='chunk duration in s for chunk processing (default 0)')
@click.option('--noise-seed', '-nseed', default=None, type=int,
              help='random seed for noise')
@click.option('--st-seed', '-stseed', default=None, type=int,
              help='random seed for spike trains')
@click.option('--temp-seed', '-tseed', default=None, type=int,
              help='random seed for template selection')
@click.option('--no-filt', is_flag=True,
              help='if True no filter is applied')
@click.option('--overlap', is_flag=True,
              help='if True it annotates overlapping spikes')
def gen_recordings(params, **kwargs):
    """Generates recordings from TEMPLATES and SPIKETRAINS"""
    # Retrieve default_params file
    this_dir, this_filename = os.path.split(__file__)
    info, config_folder = getDefaultConfig()

    if params is None:
        with open(info['recordings_params'], 'r') as pf:
            params_dict = yaml.load(pf)
    else:
        with open(params, 'r') as pf:
            params_dict = yaml.load(pf)

    if kwargs['default'] is True:
        pprint.pprint(params_dict)
        return

    if kwargs['folder'] is not None:
        params_dict['recordings'].update({'recordings_folder': kwargs['folder']})
    else:
        params_dict['recordings'].update({'recordings_folder': info['recordings_folder']})
    recordings_folder = params_dict['recordings']['recordings_folder']

    if kwargs['templates'] is None:
        print('Provide eap templates path')
        return
    else:
        if os.path.isdir(kwargs['templates']):
            params_dict['templates'].update({'templates_folder': kwargs['templates']})
        else:
            raise AttributeError("'templates' is not a folder")

    if kwargs['n_exc'] is not None:
        params_dict['spiketrains']['n_exc'] = kwargs['n_exc']
    if kwargs['n_inh'] is not None:
        params_dict['spiketrains']['n_inh'] = kwargs['n_inh']
    if kwargs['f_exc'] is not None:
        params_dict['spiketrains']['f_exc'] = kwargs['f_exc']
    if kwargs['f_inh'] is not None:
        params_dict['spiketrains']['f_inh'] = kwargs['f_inh']
    if kwargs['st_exc'] is not None:
        params_dict['spiketrains']['st_exc'] = kwargs['st_exc']
    if kwargs['st_inh'] is not None:
        params_dict['spiketrains']['st_inh'] = kwargs['st_inh']
    if kwargs['min_rate'] is not None:
        params_dict['spiketrains']['min_rate'] = kwargs['min_rate']
    if kwargs['ref_per'] is not None:
        params_dict['spiketrains']['ref_per'] = kwargs['ref_per']
    if kwargs['process'] is not None:
        params_dict['spiketrains']['process'] = kwargs['process']
    if kwargs['min_rate'] is not None:
        params_dict['spiketrains']['min_rate'] = kwargs['min_rate']
    if kwargs['duration'] is not None:
        params_dict['spiketrains']['duration'] = kwargs['duration']
    if kwargs['tstart'] is not None:
        params_dict['spiketrains']['t_start'] = kwargs['t_start']
    if kwargs['st_seed'] is not None:
        params_dict['spiketrains']['seed'] = kwargs['seed']

    if kwargs['min_dist'] is not None:
        params_dict['templates']['min_dist'] = kwargs['min_dist']
    if kwargs['min_amp'] is not None:
        params_dict['templates']['min_amp'] = kwargs['min_amp']
    if kwargs['temp_seed'] is not None:
        params_dict['templates']['seed'] = kwargs['seed']

    if kwargs['noise_lev'] is not None:
        params_dict['recordings']['noise_lev'] = kwargs['noise_lev']
    if kwargs['modulation'] is not None:
        params_dict['recordings']['modulation'] = kwargs['modulation']
    if kwargs['chunk'] is not None:
        params_dict['recordings']['chunk_duration'] = kwargs['chunk']
    if kwargs['no_filt'] is True:
        params_dict['recordings']['filter'] = False
    if kwargs['fs'] is not None:
        params_dict['recordings']['fs'] = kwargs['fs']
    else:
        params_dict['recordings']['fs'] = None
    if kwargs['noise_seed'] is not None:
        params_dict['recordings']['seed'] = kwargs['seed']

    recgen = generators.gen_recordings(templates=kwargs['templates'], params=params_dict)
    info = recgen.info

    n_neurons = info['recordings']['n_neurons']
    electrode_name = info['recordings']['electrode_name']
    duration = info['recordings']['duration']
    noise_level = info['recordings']['noise_level']

    if kwargs['fname'] is None:
        fname = 'recordings_%dcells_%s_%s_%.1fuV_%s' % (n_neurons, electrode_name, duration,
                                                        noise_level, time.strftime("%d-%m-%Y:%H:%M"))
    else:
        fname = kwargs['fname']

    rec_path = join(recordings_folder, fname)
    if not os.path.isdir(rec_path):
        os.makedirs(rec_path)

    np.save(join(rec_path, 'recordings'), recgen.recordings)
    np.save(join(rec_path, 'times'), recgen.times)
    np.save(join(rec_path, 'positions'), recgen.positions)
    np.save(join(rec_path, 'templates'), recgen.templates)
    np.save(join(rec_path, 'spiketrains'), recgen.spiketrains)
    np.save(join(rec_path, 'sources'), recgen.sources)
    np.save(join(rec_path, 'peaks'), recgen.peaks)

    with open(join(rec_path, 'info.yaml'), 'w') as f:
        yaml.dump(info, f, default_flow_style=False)

    print('\nSaved recordings in', rec_path, '\n')


@cli.command()
@click.argument('foldername')
@click.argument('h5file')
def temptohdf5(foldername, h5file):
    """Convert templates to hdf5"""
    if not h5file.endswith('.h5') and not h5file.endswith('.hdf5'):
        h5file = h5file + '.h5'
    templates_to_hdf5(foldername, h5file)
    print("Saved: ", h5file)

@cli.command()
@click.argument('h5file')
@click.argument('foldername')
def tempfromhdf5(h5file, foldername):
    """Convert templates from hdf5"""
    if not h5file.endswith('.h5') and not h5file.endswith('.hdf5'):
        raise AttributeError("'h5file' is not an hdf5 file")
    hdf5_to_templates(h5file, foldername)
    print("Saved: ", foldername)

@cli.command()
@click.argument('foldername')
@click.argument('h5file')
def rectohdf5(foldername, h5file):
    """Convert recordings to hdf5"""
    if not h5file.endswith('.h5') and not h5file.endswith('.hdf5'):
        h5file = h5file + '.h5'
    recordings_to_hdf5(foldername, h5file)
    print("Saved: ", h5file)

@cli.command()
@click.argument('h5file')
@click.argument('foldername')
def recfromhdf5(h5file, foldername):
    """Convert recordings from hdf5"""
    if not h5file.endswith('.h5') and not h5file.endswith('.hdf5'):
        raise AttributeError("'h5file' is not an hdf5 file")
    hdf5_to_recordings(h5file, foldername)
    print("Saved: ", foldername)

@cli.command()
def default_config():
    """Print default configurations"""
    info, config = getDefaultConfig()
    import pprint
    pprint.pprint(info)

@cli.command()
@click.argument('cell-models-folder')
def set_cell_models_folder(cell_models_folder):
    """Set default cell_models folder"""
    info, config = getDefaultConfig()
    if os.path.isdir(cell_models_folder):
        info['cell_models_folder'] = os.path.abspath(cell_models_folder)
        with open(join(config, 'mearec.conf'), 'w') as f:
            yaml.dump(info, f)
        print('Set default cell_models_folder to: ', cell_models_folder)
    else:
        print(cell_models_folder, ' is not a folder!')

@cli.command()
@click.argument('templates-folder')
@click.option('--create', is_flag=True, help='if True it creates the directory')
def set_templates_folder(templates_folder, create):
    """Set default templates output folder"""
    info, config = getDefaultConfig()
    templates_folder = os.path.abspath(templates_folder)
    if os.path.isdir(templates_folder):
        info['templates_folder'] = templates_folder
        with open(join(config, 'mearec.conf'), 'w') as f:
            yaml.dump(info, f)
        print('Set default templates_folder to: ', templates_folder)
    elif create:
        os.makedirs(templates_folder)
        info['templates_folder'] = templates_folder
        with open(join(config, 'mearec.conf'), 'w') as f:
            yaml.dump(info, f)
        print('Set default templates_folder to: ', templates_folder)
    else:
        print(templates_folder, ' is not a folder!')

@cli.command()
@click.argument('recordings-folder')
@click.option('--create', is_flag=True, help='if True it creates the directory')
def set_recordings_folder(recordings_folder, create):
    """Set default recordings output folder"""
    info, config = getDefaultConfig()
    recordings_folder = os.path.abspath(recordings_folder)
    if os.path.isdir(recordings_folder):
        info['recordings_folder'] = recordings_folder
        with open(join(config, 'mearec.conf'), 'w') as f:
            yaml.dump(info, f)
        print('Set default recordings_folder to: ', recordings_folder)
    elif create:
        os.makedirs(recordings_folder)
        info['recordings_folder'] = recordings_folder
        with open(join(config, 'mearec.conf'), 'w') as f:
            yaml.dump(info, f)
        print('Set default recordings_folder to: ', recordings_folder)
    else:
        print(recordings_folder, ' is not a folder!')


@cli.command()
@click.argument('templates-params')
def set_templates_params(templates_params):
    """Set default templates output folder"""
    info, config = getDefaultConfig()
    templates_params = os.path.abspath(templates_params)
    if os.path.isfile(templates_params) and (templates_params.endswith('yaml') or templates_params.endswith('yml')):
        info['templates_params'] = templates_params
        with open(join(config, 'mearec.conf'), 'w') as f:
            yaml.dump(info, f)
        print('Set default templates_params to: ', templates_params)
    else:
        print(templates_params, ' is not a yaml!')


@cli.command()
@click.argument('recordings-params')
def set_recordings_params(recordings_params):
    """Set default templates output folder"""
    info, config = getDefaultConfig()
    recordings_params = os.path.abspath(recordings_params)
    if os.path.isfile(recordings_params) and (recordings_params.endswith('yaml') or recordings_params.endswith('yml')):
        info['recordings_params'] = recordings_params
        with open(join(config, 'mearec.conf'), 'w') as f:
            yaml.dump(info, f)
        print('Set default recordings_params to: ', recordings_params)
    else:
        print(recordings_params, ' is not a yaml!')




import numpy as np
import quantities as pq
from quantities import Quantity
import yaml
import json
import neo
import elephant
import os
from os.path import join
import MEAutility as MEA
import h5py

### LOAD FUNCTIONS ###

def load_tmp_eap(templates_folder, celltypes=None, samples_per_cat=None, verbose=False):
    '''
    Loads EAP from temporary folder

    Parameters
    ----------
    templates_folder: temporary folder
    celltypes: (optional) list of celltypes to be loaded
    samples_per_cat (optional) number of eap to load per category

    Returns
    -------
    templates, locations, rotations, celltypes

    '''
    if verbose:
        print("Loading spike data ...")
    spikelist = [f for f in os.listdir(templates_folder) if f.startswith('eap')]
    loclist = [f for f in os.listdir(templates_folder) if f.startswith('pos')]
    rotlist = [f for f in os.listdir(templates_folder) if f.startswith('rot')]

    spikes_list = []
    loc_list = []
    rot_list = []
    cat_list = []

    spikelist = sorted(spikelist)
    loclist = sorted(loclist)
    rotlist = sorted(rotlist)

    loaded_categories = set()
    ignored_categories = set()

    for idx, f in enumerate(spikelist):
        celltype = f.split('-')[1][:-4]
        if verbose:
            print('loading cell type: ', f)
        if celltypes is not None:
            if celltype in celltypes:
                spikes = np.load(join(templates_folder, f))
                locs = np.load(join(templates_folder, loclist[idx]))
                rots = np.load(join(templates_folder, rotlist[idx]))

                if samples_per_cat is None or samples_per_cat > len(spikes):
                    samples_to_read = len(spikes)
                else:
                    samples_to_read = samples_per_cat

                spikes_list.extend(spikes[:samples_to_read])
                rot_list.extend(rots[:samples_to_read])
                loc_list.extend(locs[:samples_to_read])
                cat_list.extend([celltype] * samples_to_read)
                loaded_categories.add(celltype)
            else:
                ignored_categories.add(celltype)
        else:
            spikes = np.load(join(templates_folder, f))
            locs = np.load(join(templates_folder, loclist[idx]))
            rots = np.load(join(templates_folder, rotlist[idx]))

            if samples_per_cat is None or samples_per_cat > len(spikes):
                samples_to_read = len(spikes)
            else:
                samples_to_read = samples_per_cat

            spikes_list.extend(spikes[:samples_to_read])
            rot_list.extend(rots[:samples_to_read])
            loc_list.extend(locs[:samples_to_read])
            cat_list.extend([celltype] * samples_to_read)
            loaded_categories.add(celltype)

    if verbose:
        print("Done loading spike data ...")
    return np.array(spikes_list), np.array(loc_list), np.array(rot_list), np.array(cat_list, dtype=str)


def load_templates(templates, verbose=False):
    '''
    Load generated eap templates (from template_gen.py)

    Parameters
    ----------
    template_folder: templates folder

    Returns
    -------
    templates, locations, rotations, celltypes - np.arrays
    info - dict

    '''
    from MEArec import TemplateGenerator
    if verbose:
        print("Loading templates...")

    temp_dict = {}

    if os.path.isdir(templates):
        templates_folder = templates
        if os.path.isfile(join(templates_folder, 'templates.npy')):
            templates = np.load(join(templates_folder, 'templates.npy'))
            temp_dict.update({'templates': templates})
        if os.path.isfile(join(templates_folder, 'locations.npy')):
            locations = np.load(join(templates_folder, 'locations.npy'))
            temp_dict.update({'locations': locations})
        if os.path.isfile(join(templates_folder, 'rotations.npy')):
            rotations = np.load(join(templates_folder, 'rotations.npy'))
            temp_dict.update({'rotations': rotations})
        if os.path.isfile(join(templates_folder, 'celltypes.npy')):
            celltypes = np.load(join(templates_folder, 'celltypes.npy'))
            temp_dict.update({'celltypes': celltypes})
        with open(join(templates_folder, 'info.yaml'), 'r') as f:
            info = yaml.load(f)
    elif templates.endswith('h5') or templates.endswith('hdf5'):
        with h5py.File(templates, 'r') as F:
            info = json.loads(str(F['info'][()]))
            celltypes = np.array(F.get('celltypes'))
            temp_dict['celltypes'] = np.array([c.decode('utf-8') for c in celltypes])
            temp_dict['locations'] = np.array(F.get('locations'))
            temp_dict['rotations'] = np.array(F.get('rotations'))
            temp_dict['templates'] = np.array(F.get('templates'))

    if verbose:
        print("Done loading templates...")

    tempgen = TemplateGenerator(temp_dict=temp_dict, info=info)

    return tempgen


def load_recordings(recordings, verbose=False):
    '''
    Load generated recordings (from template_gen.py)

    Parameters
    ----------
    recording_folder: recordings folder

    Returns
    -------
    recordings, times, positions, templates, spiketrains, sources, peaks - np.arrays
    info - dict

    '''
    from MEArec import RecordingGenerator
    if verbose:
        print("Loading recordings...")

    rec_dict = {}

    if os.path.isdir(recordings):
        recording_folder = recordings
        if os.path.isfile(join(recording_folder, 'recordings.npy')):
            recordings = np.load(join(recording_folder, 'recordings.npy'))
            rec_dict.update({'recordings': recordings})
        if os.path.isfile(join(recording_folder, 'positions.npy')):
            positions = np.load(join(recording_folder, 'positions.npy'))
            rec_dict.update({'positions': positions})
        if os.path.isfile(join(recording_folder, 'times.npy')):
            times = np.load(join(recording_folder, 'times.npy'))
            rec_dict.update({'times': times})
        if os.path.isfile(join(recording_folder, 'templates.npy')):
            templates = np.load(join(recording_folder, 'templates.npy'))
            rec_dict.update({'templates': templates})
        if os.path.isfile(join(recording_folder, 'spiketrains.npy')):
            spiketrains = np.load(join(recording_folder, 'spiketrains.npy'))
            rec_dict.update({'spiketrains': spiketrains})
        if os.path.isfile(join(recording_folder, 'sources.npy')):
            sources = np.load(join(recording_folder, 'sources.npy'))
            rec_dict.update({'sources': sources})
        if os.path.isfile(join(recording_folder, 'peaks.npy')):
            peaks = np.load(join(recording_folder, 'peaks.npy'))
            rec_dict.update({'peaks': peaks})
        with open(join(recording_folder, 'info.yaml'), 'r') as f:
            info = yaml.load(f)
    elif recordings.endswith('h5') or recordings.endswith('hdf5'):
        with h5py.File(recordings, 'r') as F:
            info = json.loads(str(F['info'][()]))
            rec_dict['peaks'] = np.array(F.get('peaks'))
            rec_dict['positions'] = np.array(F.get('positions'))
            rec_dict['recordings'] = np.array(F.get('recordings'))
            rec_dict['sources'] = np.array(F.get('sources'))
            rec_dict['templates'] = np.array(F.get('templates'))
            rec_dict['times'] = np.array(F.get('times'))
            spiketrains = []
            for ii in range(info['recordings']['n_neurons']):
                times = np.array(F.get('spiketrains/{}/times'.format(ii)))
                t_stop = np.array(F.get('spiketrains/{}/t_stop'.format(ii)))
                annotations_str = str(F.get('spiketrains/{}/annotations'.format(ii))[()])
                annotations = json.loads(annotations_str)
                st = neo.core.SpikeTrain(
                    times,
                    t_stop=t_stop,
                    units=pq.s
                )
                st.annotations = annotations
                spiketrains.append(st)
            rec_dict['spiketrains'] = spiketrains

    if verbose:
        print("Done loading recordings...")

    recgen = RecordingGenerator(rec_dict=rec_dict, info=info)

    return recgen


def save_template_generator(tempgen, filename=None):
    if filename.endswith('h5') or filename.endswith('hdf5'):
        F = h5py.File(filename, 'w')
        F.create_dataset('info', data=json.dumps(tempgen.info))
        celltypes = [str(x).encode('utf-8') for x in tempgen.celltypes]
        F.create_dataset('celltypes', data=celltypes)
        F.create_dataset('locations', data=tempgen.locations)
        F.create_dataset('rotations', data=tempgen.rotations)
        F.create_dataset('templates', data=tempgen.templates)
        F.close()
        print('\nSaved template generator templates in', filename, '\n')
    elif filename is not None:
        save_folder = filename
        if not os.path.isdir(save_folder):
            os.makedirs(save_folder)
        np.save(join(save_folder, 'templates'), tempgen.templates)
        np.save(join(save_folder, 'locations'), tempgen.locations)
        np.save(join(save_folder, 'rotations'), tempgen.rotations)
        np.save(join(save_folder, 'celltypes'), tempgen.celltypes)
        info = tempgen.info
        yaml.dump(info, open(join(save_folder, 'info.yaml'), 'w'), default_flow_style=False)
        print('\nSaved template generator templates in', save_folder, ' folder\n')
    else:
        rot = tempgen.info['params']['rot']
        n = tempgen.info['params']['n']
        probe = tempgen.info['params']['probe']
        fname = 'templates_%d_%s_%s' % (n, probe, time.strftime("%d-%m-%Y"))
        save_folder = join(os.getcwd(), fname)
        if not os.path.isdir(save_folder):
            os.makedirs(save_folder)
        np.save(join(save_folder, 'templates'), tempgen.templates)
        np.save(join(save_folder, 'locations'), tempgen.locations)
        np.save(join(save_folder, 'rotations'), tempgen.rotations)
        np.save(join(save_folder, 'celltypes'), tempgen.celltypes)
        info = tempgen.info
        yaml.dump(info, open(join(save_folder, 'info.yaml'), 'w'), default_flow_style=False)
        print('\nSaved template generator templates in', save_folder, '\n')


def save_recording_generator(recgen, filename=None):
    if filename.endswith('h5') or filename.endswith('hdf5'):
        F = h5py.File(filename, 'w')
        F.create_dataset('info', data=json.dumps(recgen.info))
        F.create_dataset('peaks', data=recgen.peaks)
        F.create_dataset('positions', data=recgen.positions)
        F.create_dataset('recordings', data=recgen.recordings)
        F.create_dataset('sources', data=recgen.sources)
        for ii in range(len(recgen.spiketrains)):
            st = recgen.spiketrains[ii]
            F.create_dataset('spiketrains/{}/times'.format(ii), data=st.times.rescale('s').magnitude)
            F.create_dataset('spiketrains/{}/t_stop'.format(ii), data=st.t_stop)
            annotations_no_pq = {}
            for k, v in st.annotations.items():
                if isinstance(v, pq.Quantity):
                    annotations_no_pq[k] = float(v.magnitude)
                elif isinstance(v, np.ndarray):
                    annotations_no_pq[k] = list(v)
                else:
                    annotations_no_pq[k] = str(v)
            annotations_str = json.dumps(annotations_no_pq)
            F.create_dataset('spiketrains/{}/annotations'.format(ii), data=annotations_str)
        F.create_dataset('templates', data=recgen.templates)
        F.create_dataset('times', data=recgen.templates)
        F.close()
        print('\nSaved recordings in', filename, '\n')
    elif filename is not None:
        save_folder = filename
        if not os.path.isdir(save_folder):
            os.makedirs(save_folder)
        np.save(join(save_folder, 'recordings'), recgen.recordings)
        np.save(join(save_folder, 'times'), recgen.times)
        np.save(join(save_folder, 'positions'), recgen.positions)
        np.save(join(save_folder, 'templates'), recgen.templates)
        np.save(join(save_folder, 'spiketrains'), recgen.spiketrains)
        np.save(join(save_folder, 'sources'), recgen.sources)
        np.save(join(save_folder, 'peaks'), recgen.peaks)
        with open(join(save_folder, 'info.yaml'), 'w') as f:
            yaml.dump(recgen.info, f, default_flow_style=False)
        print('\nSaved recordings in', save_folder, ' folder\n')
    else:
        info = recgen.info
        n_neurons = info['recordings']['n_neurons']
        electrode_name = info['recordings']['electrode_name']
        duration = info['recordings']['duration']
        noise_level = info['recordings']['noise_level']
        fname = 'recordings_%dcells_%s_%s_%.1fuV_%s' % (n_neurons, electrode_name, duration,
                                                        noise_level, time.strftime("%d-%m-%Y:%H:%M"))
        save_folder = fname
        if not os.path.isdir(save_folder):
            os.makedirs(save_folder)
        np.save(join(save_folder, 'recordings'), recgen.recordings)
        np.save(join(save_folder, 'times'), recgen.times)
        np.save(join(save_folder, 'positions'), recgen.positions)
        np.save(join(save_folder, 'templates'), recgen.templates)
        np.save(join(save_folder, 'spiketrains'), recgen.spiketrains)
        np.save(join(save_folder, 'sources'), recgen.sources)
        np.save(join(save_folder, 'peaks'), recgen.peaks)
        with open(join(save_folder, 'info.yaml'), 'w') as f:
            yaml.dump(info, f, default_flow_style=False)
        print('\nSaved recordings in', save_folder, ' folder\n')

### H5 TOOLS ###

def hdf5_to_recordings(input_file, output_folder):
  if os.path.exists(output_folder):
    raise Exception('Output folder already exists: ' + output_folder)

  os.mkdir(output_folder)

  with h5py.File(input_file,'r') as F:
    info=json.loads(str(F['info'][()]))
    with open(output_folder+'/info.yaml','w') as f:
      yaml.dump(info,f,default_flow_style=False)

    peaks=np.array(F.get('peaks'))
    np.save(output_folder+'/peaks.npy',peaks)
    positions=np.array(F.get('positions'))
    np.save(output_folder+'/positions.npy',positions)
    recordings=np.array(F.get('recordings'))
    np.save(output_folder+'/recordings.npy',recordings)
    sources=np.array(F.get('sources'))
    np.save(output_folder+'/sources.npy',sources)
    templates=np.array(F.get('templates'))
    np.save(output_folder+'/templates.npy',templates)
    times=np.array(F.get('times'))
    np.save(output_folder+'/times.npy',times)
    spiketrains=[]
    for ii in range(info['recordings']['n_neurons']):
      times=np.array(F.get('spiketrains/{}/times'.format(ii)))
      t_stop=np.array(F.get('spiketrains/{}/t_stop'.format(ii)))
      annotations_str=str(F.get('spiketrains/{}/annotations'.format(ii))[()])
      annotations=json.loads(annotations_str)
      st=neo.core.SpikeTrain(
        times,
        t_stop=t_stop,
        units=pq.s
      )
      st.annotations=annotations
      spiketrains.append(st)
    np.save(output_folder+'/spiketrains.npy',spiketrains)


def hdf5_to_templates(input_file, output_folder):
  if os.path.exists(output_folder):
    raise Exception('Output folder already exists: '+output_folder)

  os.mkdir(output_folder)

  with h5py.File(input_file,'r') as F:
    info=json.loads(str(F['info'][()]))
    with open(output_folder+'/info.yaml','w') as f:
      yaml.dump(info,f,default_flow_style=False)

    celltypes=np.array(F.get('celltypes'))
    np.save(output_folder+'/celltypes.npy',celltypes)
    locations=np.array(F.get('locations'))
    np.save(output_folder+'/locations.npy',locations)
    rotations=np.array(F.get('rotations'))
    np.save(output_folder+'/rotations.npy',rotations)
    templates=np.array(F.get('templates'))
    np.save(output_folder+'/templates.npy',templates)


def recordings_to_hdf5(recording_folder, output_fname):
    F = h5py.File(output_fname, 'w')

    with open(recording_folder + '/info.yaml', 'r') as f:
        info = yaml.load(f)

    F.create_dataset('info', data=json.dumps(info))

    peaks = np.load(recording_folder + '/peaks.npy')
    F.create_dataset('peaks', data=peaks)
    positions = np.load(recording_folder + '/positions.npy')
    F.create_dataset('positions', data=positions)
    recordings = np.load(recording_folder + '/recordings.npy')
    F.create_dataset('recordings', data=recordings)
    sources = np.load(recording_folder + '/sources.npy')
    F.create_dataset('sources', data=sources)
    spiketrains = np.load(recording_folder + '/spiketrains.npy')
    for ii in range(len(spiketrains)):
        st = spiketrains[ii]
        F.create_dataset('spiketrains/{}/times'.format(ii), data=st.times.rescale('s').magnitude)
        F.create_dataset('spiketrains/{}/t_stop'.format(ii), data=st.t_stop)
        annotations_no_pq = {}
        for k, v in st.annotations.items():
            if isinstance(v, pq.Quantity):
                annotations_no_pq[k] = float(v.magnitude)
            elif isinstance(v, np.ndarray):
                annotations_no_pq[k] = list(v)
            else:
                annotations_no_pq[k] = v
        annotations_str = json.dumps(annotations_no_pq)
        F.create_dataset('spiketrains/{}/annotations'.format(ii), data=annotations_str)
    templates = np.load(recording_folder + '/templates.npy')
    F.create_dataset('templates', data=templates)
    templates = np.load(recording_folder + '/times.npy')
    F.create_dataset('times', data=templates)
    F.close()


def templates_to_hdf5(templates_folder, output_fname):
    F = h5py.File(output_fname, 'w')

    with open(templates_folder + '/info.yaml', 'r') as f:
        info = yaml.load(f)

    F.create_dataset('info', data=json.dumps(info))

    celltypes = np.load(templates_folder + '/celltypes.npy')
    celltypes = [str(x).encode('utf-8') for x in celltypes]
    F.create_dataset('celltypes', data=celltypes)
    locations = np.load(templates_folder + '/locations.npy')
    F.create_dataset('locations', data=locations)
    rotations = np.load(templates_folder + '/rotations.npy')
    F.create_dataset('rotations', data=rotations)
    templates = np.load(templates_folder + '/templates.npy')
    F.create_dataset('templates', data=templates)

    F.close()


### TEMPLATES INFO ###

def get_binary_cat(celltypes, excit, inhib):
    '''

    Parameters
    ----------
    celltypes: np.array with celltypes
    excit: list of excitatory celltypes
    inhib: list of inhibitory celltypes

    Returns
    -------
    bin_cat: binary celltype (E-I) - np.array

    '''
    binary_cat = []
    for i, cat in enumerate(celltypes):
        if np.any([ex in str(cat) for ex in excit]):
            binary_cat.append('E')
        elif np.any([inh in str(cat) for inh in inhib]):
            binary_cat.append('I')
    return np.array(binary_cat, dtype=str)


def get_EAP_features(EAP, feat_list, dt=None, EAP_times=None, threshold_detect=0, normalize=False):
    '''

    Parameters
    ----------
    EAP
    feat_list
    dt
    EAP_times
    threshold_detect
    normalize

    Returns
    -------

    '''
    reference_mode = 't0'
    if EAP_times is not None and dt is not None:
        test_dt = (EAP_times[-1] - EAP_times[0]) / (len(EAP_times) - 1)
        if dt != test_dt:
            raise ValueError('EAP_times and dt do not match.')
    elif EAP_times is not None:
        dt = (EAP_times[-1] - EAP_times[0]) / (len(EAP_times) - 1)
    elif dt is not None:
        EAP_times = np.arange(EAP.shape[-1]) * dt
    else:
        raise NotImplementedError('Please, specify either dt or EAP_times.')

    if len(EAP.shape) == 1:
        EAP = np.reshape(EAP, [1, 1, -1])
    elif len(EAP.shape) == 2:
        EAP = np.reshape(EAP, [1, EAP.shape[0], EAP.shape[1]])
    if len(EAP.shape) != 3:
        raise ValueError('Cannot handle EAPs with shape', EAP.shape)

    if normalize:
        signs = np.sign(np.min(EAP.reshape([EAP.shape[0], -1]), axis=1))
        norm = np.abs(np.min(EAP.reshape([EAP.shape[0], -1]), axis=1))
        EAP = np.array([EAP[i] / n if signs[i] > 0 else EAP[i] / n - 2. for i, n in enumerate(norm)])

    features = {}

    amps = np.zeros((EAP.shape[0], EAP.shape[1]))
    na_peak = np.zeros((EAP.shape[0], EAP.shape[1]))
    rep_peak = np.zeros((EAP.shape[0], EAP.shape[1]))
    if 'W' in feat_list:
        features['widths'] = np.zeros((EAP.shape[0], EAP.shape[1]))
    if 'F' in feat_list:
        features['fwhm'] = np.zeros((EAP.shape[0], EAP.shape[1]))
    if 'R' in feat_list:
        features['ratio'] = np.zeros((EAP.shape[0], EAP.shape[1]))
    if 'S' in feat_list:
        features['speed'] = np.zeros((EAP.shape[0], EAP.shape[1]))
    if 'Aids' in feat_list:
        features['Aids'] = np.zeros((EAP.shape[0], EAP.shape[1], 2), dtype=int)
    if 'Fids' in feat_list:
        features['Fids'] = []
    if 'FV' in feat_list:
        features['fwhm_V'] = np.zeros((EAP.shape[0], EAP.shape[1]))
    if 'Na' in feat_list:
        features['na'] = np.zeros((EAP.shape[0], EAP.shape[1]))
    if 'Rep' in feat_list:
        features['rep'] = np.zeros((EAP.shape[0], EAP.shape[1]))

    for i in range(EAP.shape[0]):
        # For AMP feature
        min_idx = np.array([np.unravel_index(EAP[i, e].argmin(), EAP[i, e].shape)[0] for e in
                            range(EAP.shape[1])])
        max_idx = np.array([np.unravel_index(EAP[i, e, min_idx[e]:].argmax(),
                                             EAP[i, e, min_idx[e]:].shape)[0]
                            + min_idx[e] for e in range(EAP.shape[1])])
        # for na and rep
        min_elid, min_idx_na = np.unravel_index(EAP[i].argmin(), EAP[i].shape)
        max_idx_rep = EAP[i, min_elid, min_idx_na:].argmax() + min_idx_na
        na_peak[i, :] = EAP[i, :, min_idx_na]
        rep_peak[i, :] = EAP[i, :, max_idx_rep]

        if 'Aids' in feat_list:
            features['Aids'][i] = np.vstack((min_idx, max_idx)).T

        amps[i, :] = np.array([EAP[i, e, max_idx[e]] - EAP[i, e, min_idx[e]] for e in range(EAP.shape[1])])
        # If below 'detectable threshold, set amp and width to 0
        if normalize:
            too_low = np.where(amps[i, :] < threshold_detect / norm[i])
        else:
            too_low = np.where(amps[i, :] < threshold_detect)
        amps[i, too_low] = 0

        if 'R' in feat_list:
            min_id_ratio = np.array([np.unravel_index(EAP[i, e, min_idx_na:].argmin(),
                                                      EAP[i, e, min_idx_na:].shape)[0]
                                     + min_idx_na for e in range(EAP.shape[1])])
            max_id_ratio = np.array([np.unravel_index(EAP[i, e, min_idx_na:].argmax(),
                                                      EAP[i, e, min_idx_na:].shape)[0]
                                     + min_idx_na for e in range(EAP.shape[1])])
            features['ratio'][i, :] = np.array([np.abs(EAP[i, e, max_id_ratio[e]]) /
                                                np.abs(EAP[i, e, min_id_ratio[e]])
                                                for e in range(EAP.shape[1])])
            # If below 'detectable threshold, set amp and width to 0
            too_low = np.where(amps[i, :] < threshold_detect)
            features['ratio'][i, too_low] = 1
        if 'S' in feat_list:
            features['speed'][i, :] = np.array((min_idx - min_idx_na) * dt)
            features['speed'][i, too_low] = min_idx_na * dt

        if 'W' in feat_list:
            features['widths'][i, :] = np.abs(EAP_times[max_idx] - EAP_times[min_idx])
            features['widths'][i, too_low] = EAP.shape[2] * dt  # EAP_times[-1]-EAP_times[0]

        if 'F' in feat_list:
            min_peak = np.min(EAP[i], axis=1)
            if reference_mode == 't0':
                # reference voltage is zeroth voltage entry
                fwhm_ref = np.array([EAP[i, e, 0] for e in range(EAP.shape[1])])
            elif reference_mode == 'maxd2EAP':
                # reference voltage is taken at peak onset
                # peak onset is defined as id of maximum 2nd derivative of EAP
                peak_onset = np.array([np.argmax(savitzky_golay(EAP[i, e], 5, 2, deriv=2)[:min_idx[e]])
                                       for e in range(EAP.shape[1])])
                fwhm_ref = np.array([EAP[i, e, peak_onset[e]] for e in range(EAP.shape[1])])
            else:
                raise NotImplementedError('Reference mode ' + reference_mode + ' for FWHM calculation not implemented.')
            fwhm_V = (fwhm_ref + min_peak) / 2.
            id_trough = [np.where(EAP[i, e] < fwhm_V[e])[0] for e in range(EAP.shape[1])]
            if 'Fids' in feat_list:
                features['Fids'].append(id_trough)
            if 'FV' in feat_list:
                features['fwhm_V'][i, :] = fwhm_V

            # linear interpolation due to little number of data points during peak

            # features['fwhm'][i,:] = np.array([(len(t)+1)*dt+(fwhm_V[e]-EAP[i,e,t[0]-1])/(EAP[i,e,t[0]]-EAP[i,e,t[0]-1])*dt -(fwhm_V[e]-EAP[i,e,t[-1]])/(EAP[i,e,min(t[-1]+1,EAP.shape[2]-1)]-EAP[i,e,t[-1]])*dt if len(t)>0 else np.infty for e,t in enumerate(id_trough)])

            # no linear interpolation
            features['fwhm'][i, :] = [(id_trough[e][-1] - id_trough[e][0]) * dt if len(id_trough[e]) > 1 \
                                          else EAP.shape[2] * dt for e in range(EAP.shape[1])]
            features['fwhm'][i, too_low] = EAP.shape[2] * dt  # EAP_times[-1]-EAP_times[0]

    if 'A' in feat_list:
        features.update({'amps': amps})
    if 'Na' in feat_list:
        features.update({'na': na_peak})
    if 'Rep' in feat_list:
        features.update({'rep': rep_peak})

    return features


### TEMPLATES OPERATIONS ###

def is_position_within_boundaries(position, x_lim, y_lim, z_lim):
    '''

    Parameters
    ----------
    position
    x_lim
    y_lim
    z_lim

    Returns
    -------

    '''
    valid_position = True
    if x_lim is not None:
        if position[0] < x_lim[0] or position[0] > x_lim[1]:
            valid_position = False
    if y_lim is not None:
        if position[1] < y_lim[0] or position[1] > y_lim[1]:
            valid_position = False
    if z_lim is not None:
        if position[2] < z_lim[0] or position[2] > z_lim[1]:
            valid_position = False
    return valid_position


def select_templates(loc, spikes, bin_cat, n_exc, n_inh, min_dist=25, x_lim=None, y_lim=None, z_lim=None,
                     min_amp=None, drift=False, drift_dir_ang=[], preferred_dir=None, ang_tol=30, verbose=False):
    '''

    Parameters
    ----------
    loc
    spikes
    bin_cat
    n_exc
    n_inh
    min_dist
    x_lim
    y_lim
    z_lim
    min_amp
    drift
    drift_dir_ang
    preferred_dir
    ang_tol
    verbose

    Returns
    -------

    '''
    pos_sel = []
    idxs_sel = []
    categories = np.unique(bin_cat)
    if 'E' in categories and 'I' in categories:
        if verbose:
            print('Selecting Excitatory and Inhibitory cells')
        excinh = True
    else:
        if verbose:
            print('Selecting random templates (cell types not specified)')
        excinh = False
    permuted_idxs = np.random.permutation(len(bin_cat))
    permuted_bin_cats = bin_cat[permuted_idxs]

    if not min_amp:
        min_amp = 0

    if drift:
        if len(drift_dir_ang) == 0 or preferred_dir == None:
            raise Exception('For drift selection provide drifting angles and preferred drift direction')

    placed_exc = 0
    placed_inh = 0
    n_sel = 0
    n_sel_exc = 0
    n_sel_inh = 0
    iter = 0

    for i, (id_cell, bcat) in enumerate(zip(permuted_idxs, permuted_bin_cats)):
        placed = False
        iter += 1
        if n_sel == n_exc + n_inh:
            break
        if excinh:
            if bcat == 'E':
                if n_sel_exc < n_exc:
                    dist = np.array([np.linalg.norm(loc[id_cell] - p) for p in pos_sel])
                    if np.any(dist < min_dist):
                        if verbose:
                            print('Distance violation', dist, iter)
                        pass
                    else:
                        amp = np.max(np.abs(spikes[id_cell]))
                        if not drift:
                            if is_position_within_boundaries(loc[id_cell], x_lim, y_lim, z_lim) and amp > min_amp:
                                    # save cell
                                    pos_sel.append(loc[id_cell])
                                    idxs_sel.append(id_cell)
                                    n_sel += 1
                                    placed = True
                            else:
                                if verbose:
                                    print('Amplitude or boundary violation', amp, loc[id_cell], iter)
                        else:
                            # drift
                            if is_position_within_boundaries(loc[id_cell], x_lim, y_lim, z_lim) and amp > min_amp:
                                    # save cell
                                if np.abs(drift_dir_ang[id_cell] - preferred_dir) < ang_tol:
                                    pos_sel.append(loc[id_cell])
                                    idxs_sel.append(id_cell)
                                    n_sel += 1
                                    placed = True
                                else:
                                    if verbose:
                                        print('Drift violation', loc[id_cell], iter)
                            else:
                                if verbose:
                                    print('Amplitude or boundary violation', amp, loc[id_cell], iter)
                    if placed:
                        n_sel_exc += 1
            elif bcat == 'I':
                if n_sel_inh < n_inh:
                    dist = np.array([np.linalg.norm(loc[id_cell] - p) for p in pos_sel])
                    if np.any(dist < min_dist):
                        if verbose:
                            print('Distance violation', dist, iter)
                        pass
                    else:
                        amp = np.max(np.abs(spikes[id_cell]))
                        if not drift:
                            if is_position_within_boundaries(loc[id_cell], x_lim, y_lim, z_lim) and amp > min_amp:
                                    # save cell
                                    pos_sel.append(loc[id_cell])
                                    idxs_sel.append(id_cell)
                                    n_sel += 1
                                    placed = True
                            else:
                                if verbose:
                                    print('Amplitude or boundary violation', amp, loc[id_cell], iter)
                        else:
                            # drift
                            if is_position_within_boundaries(loc[id_cell], x_lim, y_lim, z_lim) and amp > min_amp:
                                    # save cell
                                if np.abs(drift_dir_ang[id_cell] - preferred_dir) < ang_tol:
                                    pos_sel.append(loc[id_cell])
                                    idxs_sel.append(id_cell)
                                    n_sel += 1
                                    placed = True
                                else:
                                    if verbose:
                                        print('Drift violation', loc[id_cell], iter)
                            else:
                                if verbose:
                                    print('Amplitude or boundary violation', amp, loc[id_cell], iter)
                    if placed:
                        n_sel_inh += 1
        else:
            dist = np.array([np.linalg.norm(loc[id_cell] - p) for p in pos_sel])
            if np.any(dist < min_dist):
                if verbose:
                    print('Distance violation', dist, iter)
                pass
            else:
                amp = np.max(np.abs(spikes[id_cell]))
                if not drift:
                    if is_position_within_boundaries(loc[id_cell], x_lim, y_lim, z_lim) and amp > min_amp:
                        # save cell
                        pos_sel.append(loc[id_cell])
                        idxs_sel.append(id_cell)
                        placed = True
                    else:
                        if verbose:
                            print('Amplitude or boundary violation', amp, loc[id_cell], iter)
                else:
                    # drift
                    if is_position_within_boundaries(loc[id_cell], x_lim, y_lim, z_lim) and amp > min_amp:
                        # save cell
                        if np.abs(drift_dir_ang[id_cell] - preferred_dir) < ang_tol:
                            pos_sel.append(loc[id_cell])
                            idxs_sel.append(id_cell)
                            placed = True
                        else:
                            if verbose:
                                print('Drift violation', loc[id_cell], iter)
                    else:
                        if verbose:
                            print('Amplitude or boundary violation', amp, loc[id_cell], iter)
            if placed:
                n_sel += 1
                print(i, len(permuted_idxs), n_sel)
    if i == len(permuted_idxs)-1 and n_sel < n_exc + n_inh:
        raise RuntimeError("Templates could not be selected. \n"
                           "Decrease number of spiketrains, decrease 'min-dist', or use more templates.")

    return idxs_sel


def cubic_padding(spike, pad_len, fs, percent_mean=0.2):
    '''
    Cubic spline padding on left and right side to 0

    Parameters
    ----------
    spike
    pad_len
    fs

    Returns
    -------
    padded_template

    '''
    import scipy.interpolate as interp
    n_pre = int(pad_len[0] * fs)
    n_post = int(pad_len[1] * fs)

    padded_template = np.zeros((spike.shape[0], int(n_pre) + spike.shape[1] + n_post))
    splines = np.zeros((spike.shape[0], int(n_pre) + spike.shape[1] + n_post))

    for i, sp in enumerate(spike):
        # Remove inital offset
        padded_sp = np.zeros(n_pre + len(sp) + n_post)
        padded_t = np.arange(len(padded_sp))
        initial_offset = np.mean(sp[0])
        sp -= initial_offset

        x_pre = float(n_pre)
        x_pre_pad = np.arange(n_pre)
        x_post = float(n_post)
        x_post_pad = np.arange(n_post)[::-1]

        off_pre = sp[0]
        off_post = sp[-1]
        m_pre = sp[0] / x_pre
        m_post = sp[-1] / x_post

        padded_sp[:n_pre] = m_pre * x_pre_pad
        padded_sp[n_pre:-n_post] = sp
        padded_sp[-n_post:] = m_post * x_post_pad

        f = interp.interp1d(padded_t, padded_sp, kind='cubic')
        splines[i] = f(np.arange(len(padded_sp)))

        padded_template[i, :n_pre] = f(x_pre_pad)
        padded_template[i, n_pre:-n_post] = sp
        padded_template[i, -n_post:] = f(np.arange(n_pre + len(sp), n_pre + len(sp) + n_post))

    return padded_template, splines


def find_overlapping_templates(templates, thresh=0.7):
    '''

    Parameters
    ----------
    templates
    thresh

    Returns
    -------

    '''
    overlapping_pairs = []

    for i in range(templates.shape[0] - 1):
        temp_1 = templates[i]
        max_ptp = (np.array([np.ptp(t) for t in temp_1]).max())
        max_ptp_idx = (np.array([np.ptp(t) for t in temp_1]).argmax())

        for j in range(i + 1, templates.shape[0]):
            temp_2 = templates[j]
            ptp_on_max = np.ptp(temp_2[max_ptp_idx])

            max_ptp_2 = (np.array([np.ptp(t) for t in temp_2]).max())

            max_peak = np.max([ptp_on_max, max_ptp])
            min_peak = np.min([ptp_on_max, max_ptp])

            if min_peak > thresh * max_peak and ptp_on_max > thresh * max_ptp_2:
                overlapping_pairs.append([i, j])  # , max_ptp_idx, max_ptp, ptp_on_max

    return np.array(overlapping_pairs)


### SPIKETRAIN OPERATIONS ###

def annotate_overlapping_spikes(gtst, t_jitt = 1*pq.ms, overlapping_pairs=None, verbose=False, parallel=True):
    '''

    Parameters
    ----------
    gtst
    t_jitt
    overlapping_pairs
    verbose

    Returns
    -------

    '''
    nprocesses = len(gtst)
    if parallel:
        import multiprocessing
        # t_start = time.time()
        pool = multiprocessing.Pool(nprocesses)
        results = [pool.apply_async(annotate_parallel(i, st_i, gtst, overlapping_pairs, t_jitt, ))
                   for i, st_i in enumerate(gtst)]
    else:
        # find overlapping spikes
        for i, st_i in enumerate(gtst):
            if verbose:
                print('SPIKETRAIN ', i)
            over = np.array(['NO'] * len(st_i))
            for i_sp, t_i in enumerate(st_i):
                for j, st_j in enumerate(gtst):
                    if i != j:
                        # find overlapping
                        id_over = np.where((st_j > t_i - t_jitt) & (st_j < t_i + t_jitt))[0]
                        if not np.any(overlapping_pairs):
                            if len(id_over) != 0:
                                over[i_sp] = 'O'
                                # if verbose:
                                #     print('found overlap! spike 1: ', i, t_i, ' spike 2: ', j, st_j[id_over])
                        else:
                            pair = [i, j]
                            pair_i = [j, i]
                            if np.any([np.all(pair == p) for p in overlapping_pairs]) or \
                                    np.any([np.all(pair_i == p) for p in overlapping_pairs]):
                                if len(id_over) != 0:
                                    over[i_sp] = 'SO'
                                    # if verbose:
                                    #     print('found spatial overlap! spike 1: ', i, t_i, ' spike 2: ', j, st_j[id_over])
                            else:
                                if len(id_over) != 0:
                                    over[i_sp] = 'O'
                                    # if verbose:
                                    #     print('found overlap! spike 1: ', i, t_i, ' spike 2: ', j, st_j[id_over])
            st_i.annotate(overlap=over)


def annotate_parallel(i, st_i, gtst, overlapping_pairs, t_jitt):
    '''

    Parameters
    ----------
    i
    st_i
    gtst
    overlapping_pairs
    t_jitt

    Returns
    -------

    '''
    print('SPIKETRAIN ', i)
    over = np.array(['NO'] * len(st_i))
    for i_sp, t_i in enumerate(st_i):
        for j, st_j in enumerate(gtst):
            if i != j:
                # find overlapping
                id_over = np.where((st_j > t_i - t_jitt) & (st_j < t_i + t_jitt))[0]
                if not np.any(overlapping_pairs):
                    if len(id_over) != 0:
                        over[i_sp] = 'O'
                        # if verbose:
                        #     print('found overlap! spike 1: ', i, t_i, ' spike 2: ', j, st_j[id_over])
                else:
                    pair = [i, j]
                    pair_i = [j, i]
                    if np.any([np.all(pair == p) for p in overlapping_pairs]) or \
                            np.any([np.all(pair_i == p) for p in overlapping_pairs]):
                        if len(id_over) != 0:
                            over[i_sp] = 'SO'
                            # if verbose:
                            #     print('found spatial overlap! spike 1: ', i, t_i, ' spike 2: ', j, st_j[id_over])
                    else:
                        if len(id_over) != 0:
                            over[i_sp] = 'O'
                            # if verbose:
                            #     print('found overlap! spike 1: ', i, t_i, ' spike 2: ', j, st_j[id_over])
    st_i.annotate(overlap=over)


def resample_spiketrains(spiketrains, fs=None, T=None):
    '''
    Resamples spike trains. Provide either fs or T parameters
    Parameters
    ----------
    fs: new sampling frequency (quantity)
    T: new period (quantity)

    Returns
    -------
    matrix with resampled binned spike trains

    '''
    import elephant.conversion as conv

    resampled_mat = []
    if not fs and not T:
        print('Provide either sampling frequency fs or time period T')
        return
    elif fs:
        if not isinstance(fs, Quantity):
            raise ValueError("fs must be of type pq.Quantity")
        binsize = 1./fs
        binsize.rescale('ms')
        resampled_mat = []
        for sts in spiketrains:
            spikes = conv.BinnedSpikeTrain(sts, binsize=binsize).to_array()
            resampled_mat.append(np.squeeze(spikes))
    elif T:
        binsize = T
        if not isinstance(T, Quantity):
            raise ValueError("T must be of type pq.Quantity")
        binsize.rescale('ms')
        resampled_mat = []
        for sts in spiketrains:
            spikes = conv.BinnedSpikeTrain(sts, binsize=binsize).to_array()
            resampled_mat.append(np.squeeze(spikes))
    return np.array(resampled_mat)


### CONVOLUTION OPERATIONS ###

def ISI_amplitude_modulation(st, n_el=1, mrand=1, sdrand=0.05, n_spikes=1, exp=0.5, mem_ISI = 10*pq.ms):
    '''

    Parameters
    ----------
    st
    mrand
    sdrand
    n_spikes
    exp
    mem_ISI

    Returns
    -------

    '''

    import elephant.statistics as stat

    if n_el == 1:
        ISI = stat.isi(st).rescale('ms')
        # mem_ISI = 2*mean_ISI
        amp_mod = np.zeros(len(st))
        amp_mod[0] = sdrand*np.random.randn() + mrand
        cons = np.zeros(len(st))

        for i, isi in enumerate(ISI):
            if n_spikes == 1:
                if isi > mem_ISI:
                    amp_mod[i + 1] = sdrand * np.random.randn() + mrand
                else:
                    amp_mod[i + 1] = isi.magnitude ** exp * (1. / mem_ISI.magnitude ** exp) + sdrand * np.random.randn()
            else:
                consecutive = 0
                bursting=True
                while consecutive < n_spikes and bursting:
                    if i-consecutive >= 0:
                        if ISI[i-consecutive] > mem_ISI:
                            bursting = False
                        else:
                            consecutive += 1
                    else:
                        bursting = False

                if consecutive == 0:
                    amp_mod[i + 1] = sdrand * np.random.randn() + mrand
                elif consecutive==1:
                    amp = (isi / float(consecutive)) ** exp * (1. / mem_ISI.magnitude ** exp)
                    # scale std by amp
                    amp_mod[i + 1] = amp + amp * sdrand * np.random.randn()
                else:
                    if i != len(ISI):
                        isi_mean = np.mean(ISI[i-consecutive+1:i+1])
                    else:
                        isi_mean = np.mean(ISI[i - consecutive + 1:])
                    amp = (isi_mean/float(consecutive)) ** exp * (1. / mem_ISI.magnitude ** exp)
                    # scale std by amp
                    amp_mod[i + 1] = amp + amp * sdrand * np.random.randn()

                cons[i+1] = consecutive
    else:
        if n_spikes == 0:
            amp_mod = []
            cons = []
            for i, sp in enumerate(st):
                amp_mod.append(sdrand * np.random.randn(n_el) + mrand)

    return np.array(amp_mod), cons

# TODO use cut outs to align spikes!
def convolve_single_template(spike_id, spike_bin, template, cut_out=None, modulation=False, amp_mod=None):
    '''

    Parameters
    ----------
    spike_id
    spike_bin
    template
    modulation
    amp_mod

    Returns
    -------

    '''
    if len(template.shape) == 2:
        njitt = template.shape[0]
        len_spike = template.shape[1]
    else:
        len_spike = template.shape[0]
    spike_pos = np.where(spike_bin == 1)[0]
    n_samples = len(spike_bin)
    gt_source = np.zeros(n_samples)

    if len(template.shape) == 2:
        rand_idx = np.random.randint(njitt)
        # print('rand_idx: ', rand_idx)
        temp_jitt = template[rand_idx]
        if not modulation:
            for pos, spos in enumerate(spike_pos):
                if spos - cut_out[0] >= 0 and spos - cut_out[0] + len_spike <= n_samples:
                    gt_source[spos - cut_out[0]:spos - cut_out[0] + len_spike] +=  temp_jitt
                elif spos - cut_out[0] < 0:
                    diff = -(spos - cut_out[0])
                    gt_source[:spos - cut_out[0] + len_spike] += temp_jitt[diff:]
                else:
                    diff = n_samples - (spos - cut_out[0])
                    gt_source[spos - cut_out[0]:] += temp_jitt[:diff]
        else:
            # print('Template-Electrode modulation')
            for pos, spos in enumerate(spike_pos):
                if spos - cut_out[0] >= 0 and spos - cut_out[0] + len_spike <= n_samples:
                    gt_source[spos - cut_out[0]:spos - cut_out[0] + len_spike] += amp_mod[pos]*temp_jitt
                elif spos - cut_out[0] < 0:
                    diff = -(spos - cut_out[0])
                    gt_source[:spos - cut_out[0] + len_spike] += amp_mod[pos]*temp_jitt[diff:]
                else:
                    diff = n_samples - (spos - cut_out[0])
                    gt_source[spos - cut_out[0]:] += amp_mod[pos]*temp_jitt[:diff]
    else:
        # print('No modulation')
        for pos, spos in enumerate(spike_pos):
            if spos - cut_out[0] >= 0 and spos - cut_out[0] + len_spike <= n_samples:
                gt_source[spos - cut_out[0]:spos - cut_out[0] + len_spike] += template
            elif spos - cut_out[0] < 0:
                diff = -(spos - cut_out[0])
                gt_source[:spos - cut_out[0] + len_spike] += template[diff:]
            else:
                diff = n_samples - (spos - cut_out[0])
                gt_source[spos - cut_out[0]:] += template[:diff]

    return gt_source


def convolve_templates_spiketrains(spike_id, spike_bin, template,  cut_out=None, modulation=False, amp_mod=None, recordings=[]):
    '''

    Parameters
    ----------
    spike_id
    spike_bin
    template
    modulation
    amp_mod
    recordings

    Returns
    -------

    '''
    print('START: convolution with spike ', spike_id)
    if len(template.shape) == 3:
        njitt = template.shape[0]
        n_elec = template.shape[1]
        len_spike = template.shape[2]
    else:
        n_elec = template.shape[0]
        len_spike = template.shape[1]
    n_samples = len(spike_bin)
    if len(recordings) == 0:
        recordings = np.zeros((n_elec, n_samples))
    if cut_out is None:
        cut_out = [len_spike//2, len_spike//2]        

    # recordings_test = np.zeros((n_elec, n_samples))
    if not modulation:
        spike_pos = np.where(spike_bin == 1)[0]
        amp_mod = np.ones_like(spike_pos)
        if len(template.shape) == 3:
            rand_idx = np.random.randint(njitt)
            # print('rand_idx: ', rand_idx)
            temp_jitt = template[rand_idx]
            # print('No modulation')
            for pos, spos in enumerate(spike_pos):
                if spos - cut_out[0] >= 0 and spos + cut_out[1] <= n_samples:
                    recordings[:, spos - cut_out[0]:spos + cut_out[1]] += amp_mod[pos] * temp_jitt
                elif spos - cut_out[0] < 0:
                    diff = -(spos - cut_out[0])
                    recordings[:, :spos + cut_out[1]] += amp_mod[pos] * temp_jitt[:, diff:]
                else:
                    diff = n_samples - (spos - cut_out[0])
                    recordings[:, spos - cut_out[0]:] += amp_mod[pos] * temp_jitt[:, :diff]

            raise Exception()
        else:
            # print('No jitter')
            for pos, spos in enumerate(spike_pos):
                if spos - cut_out[0] >= 0 and spos + cut_out[1] <= n_samples:
                    recordings[:, spos - cut_out[0]:spos + cut_out[1]] += amp_mod[
                                                                                                  pos] * template
                elif spos - cut_out[0] < 0:
                    diff = -(spos - cut_out[0])
                    recordings[:, :spos + cut_out[1]] += amp_mod[pos] * template[:, diff:]
                else:
                    diff = n_samples - (spos - cut_out[0])
                    recordings[:, spos - cut_out[0]:] += amp_mod[pos] * template[:, :diff]
    else:
        assert amp_mod is not None
        spike_pos = np.where(spike_bin == 1)[0]
        if len(template.shape) == 3:
            rand_idx = np.random.randint(njitt)
            # print('rand_idx: ', rand_idx)
            temp_jitt = template[rand_idx]
            if not isinstance(amp_mod[0], (list, tuple, np.ndarray)):
                #print('Template modulation')
                for pos, spos in enumerate(spike_pos):
                    if spos - cut_out[0] >= 0 and spos + cut_out[1] <= n_samples:
                        recordings[:, spos - cut_out[0]:spos + cut_out[1]] +=  amp_mod[pos] * temp_jitt
                    elif spos - cut_out[0] < 0:
                        diff = -(spos - cut_out[0])
                        recordings[:, :spos + cut_out[1]] += amp_mod[pos] * temp_jitt[:, diff:]
                    else:
                        diff = n_samples-(spos - cut_out[0])
                        recordings[:, spos - cut_out[0]:] += amp_mod[pos] * temp_jitt[:, :diff]
            else:
                #print('Electrode modulation')
                for pos, spos in enumerate(spike_pos):
                    if spos - cut_out[0] >= 0 and spos + cut_out[1] <= n_samples:
                        recordings[:, spos - cut_out[0]:spos + cut_out[1]] += \
                            [a * t for (a, t) in zip(amp_mod[pos], temp_jitt)]
                    elif spos - cut_out[0] < 0:
                        diff = -(spos - cut_out[0])
                        recordings[:, :spos + cut_out[1]] += \
                            [a * t for (a, t) in zip(amp_mod[pos], temp_jitt[:, diff:])]
                        # recordings[:, :spos + cut_out[1]] += amp_mod[pos] * template[:, diff:]
                    else:
                        diff = n_samples-(spos - cut_out[0])
                        recordings[:, spos - cut_out[0]:] += \
                            [a * t for (a, t) in zip(amp_mod[pos], temp_jitt[:, :diff])]
        else:
            if not isinstance(amp_mod[0], (list, tuple, np.ndarray)):
                #print('Template modulation')
                for pos, spos in enumerate(spike_pos):
                    if spos - cut_out[0] >= 0 and spos + cut_out[1] <= n_samples:
                        recordings[:, spos - cut_out[0]:spos + cut_out[1]] += amp_mod[
                                                                                                      pos] * template
                    elif spos - cut_out[0] < 0:
                        diff = -(spos - cut_out[0])
                        recordings[:, :spos + cut_out[1]] += amp_mod[pos] * template[:, diff:]
                    else:
                        diff = n_samples - (spos - cut_out[0])
                        recordings[:, spos - cut_out[0]:] += amp_mod[pos] * template[:, :diff]

            else:
                #print('Electrode modulation')
                for pos, spos in enumerate(spike_pos):
                    if spos - cut_out[0] >= 0 and spos + cut_out[1] <= n_samples:
                        recordings[:, spos - cut_out[0]:spos + cut_out[1]] += \
                            [a * t for (a, t) in zip(amp_mod[pos], template)]
                    elif spos - cut_out[0] < 0:
                        diff = -(spos - cut_out[0])
                        recordings[:, : spos + cut_out[1]] += \
                            [a * t for (a, t) in zip(amp_mod[pos], template[:, diff:])]
                        # recordings[:, :spos + cut_out[1]] += amp_mod[pos] * template[:, diff:]
                    else:
                        diff = n_samples-(spos - cut_out[0])
                        recordings[:, spos - cut_out[0]:] += \
                            [a * t for (a, t) in zip(amp_mod[pos], template[:, :diff])]
                        # recordings[:, spos - cut_out[0]:] += amp_mod[pos] * template[:, :diff]


    #print('DONE: convolution with spike ', spike_id)

    return recordings


def convolve_drifting_templates_spiketrains(spike_id, spike_bin, template, fs, loc, v_drift, t_start_drift,
                                            modulation=False, amp_mod=None, recordings=[], n_step_sec=1):
    '''

    Parameters
    ----------
    spike_id
    spike_bin
    template
    fs
    loc
    v_drift
    t_start_drift
    modulation
    amp_mod
    recordings
    n_step_sec

    Returns
    -------

    '''
    print('START: convolution with spike ', spike_id)
    if len(template.shape) == 4:
        njitt = template.shape[1]
        n_elec = template.shape[2]
        len_spike = template.shape[3]
    elif len(template.shape) == 3:
        n_elec = template.shape[1]
        len_spike = template.shape[2]
    n_samples = len(spike_bin)
    n_step_sec = 1
    dur = (n_samples / fs).rescale('s').magnitude
    t_steps = np.arange(0, dur, n_step_sec)
    n_step_sample = n_step_sec * int(fs.magnitude)
    dt = 2 ** -5

    mixing = np.zeros((int(n_samples/float(fs.rescale('Hz').magnitude)), n_elec))
    if len(recordings) == 0:
        recordings = np.zeros((n_elec, n_samples))

    # recordings_test = np.zeros((n_elec, n_samples))
    if not modulation:
        spike_pos = np.where(spike_bin == 1)[0]
        amp_mod = np.ones_like(spike_pos)
        if len(template.shape) == 4:
            rand_idx = np.random.randint(njitt)
            print('rand_idx: ', rand_idx)
            print('No modulation')
            for pos, spos in enumerate(spike_pos):
                sp_time = spos / fs
                if sp_time < t_start_drift:
                    print(sp_time, 'No drift', loc[0])
                    temp_idx = 0
                    temp_jitt = template[temp_idx, rand_idx]
                else:
                    # compute current position
                    new_pos = np.array(loc[0, 1:] + v_drift * (sp_time - t_start_drift).rescale('s').magnitude)
                    temp_idx = np.argmin([np.linalg.norm(p - new_pos) for p in loc[:, 1:]])
                    print(sp_time, temp_idx, 'Drifting', new_pos, loc[temp_idx, 1:])
                    temp_jitt = template[temp_idx, rand_idx]

                if spos - cut_out[0] >= 0 and spos + cut_out[1] <= n_samples:
                    recordings[:, spos - cut_out[0]:spos + cut_out[1]] += amp_mod[pos] * temp_jitt
                elif spos - cut_out[0] < 0:
                    diff = -(spos - cut_out[0])
                    recordings[:, :spos + cut_out[1]] += amp_mod[pos] * temp_jitt[:, diff:]
                else:
                    diff = n_samples - (spos - cut_out[0])
                    recordings[:, spos - cut_out[0]:] += amp_mod[pos] * temp_jitt[:, :diff]

            for i, t in enumerate(t_steps):
                if t < t_start_drift:
                    temp_idx = 0
                    temp_jitt = template[temp_idx, rand_idx]
                else:
                    # compute current position
                    new_pos = np.array(loc[0, 1:] + v_drift * (t - t_start_drift.rescale('s').magnitude))
                    temp_idx = np.argmin([np.linalg.norm(p - new_pos) for p in loc[:, 1:]])
                    temp_jitt = template[temp_idx, rand_idx]

                feat = get_EAP_features(np.squeeze(temp_jitt), ['Na'], dt=dt)
                mixing[i] = -np.squeeze(feat['na'])
        else:
            print('No jitter')
            for pos, spos in enumerate(spike_pos):
                sp_time = spos / fs
                if sp_time < t_start_drift:
                    temp_idx = 0
                    temp = template[temp_idx]
                else:
                    # compute current position
                    new_pos = np.array(loc[0, 1:] + v_drift * (sp_time - t_start_drift).rescale('s').magnitude)
                    temp_idx = np.argmin([np.linalg.norm(p - new_pos) for p in loc[:, 1:]])
                    temp = template[temp_idx]
                if spos - cut_out[0] >= 0 and spos + cut_out[1] <= n_samples:
                    recordings[:, spos - cut_out[0]:spos + cut_out[1]] += amp_mod[pos] * temp
                elif spos - cut_out[0] < 0:
                    diff = -(spos - cut_out[0])
                    recordings[:, :spos + cut_out[1]] += amp_mod[pos] * temp[:, diff:]
                else:
                    diff = n_samples - (spos - cut_out[0])
                    recordings[:, spos - cut_out[0]:] += amp_mod[pos] * temp[:, :diff]
            for i, t in enumerate(t_steps):
                if t < t_start_drift:
                    temp_idx = 0
                    temp_jitt = template[temp_idx]
                else:
                    # compute current position
                    new_pos = np.array(loc[0, 1:] + v_drift * (t - t_start_drift.rescale('s').magnitude))
                    temp_idx = np.argmin([np.linalg.norm(p - new_pos) for p in loc[:, 1:]])
                    temp_jitt = template[temp_idx]

                feat = get_EAP_features(np.squeeze(temp_jitt), ['Na'], dt=dt)
                mixing[i] = -np.squeeze(feat['na'])
    else:
        assert amp_mod is not None
        spike_pos = np.where(spike_bin == 1)[0]
        if len(template.shape) == 4:
            rand_idx = np.random.randint(njitt)
            print('rand_idx: ', rand_idx)
            if not isinstance(amp_mod[0], (list, tuple, np.ndarray)):
                print('Template modulation')
                for pos, spos in enumerate(spike_pos):
                    sp_time = spos / fs
                    if sp_time < t_start_drift:
                        print(sp_time, 'No drift', loc[0])
                        temp_idx = 0
                        temp_jitt = template[temp_idx, rand_idx]
                    else:
                        # compute current position
                        new_pos = np.array(loc[0, 1:] + v_drift * (sp_time - t_start_drift).rescale('s').magnitude)
                        temp_idx = np.argmin([np.linalg.norm(p - new_pos) for p in loc[:, 1:]])
                        temp_jitt = template[temp_idx, rand_idx]
                        print(sp_time, temp_idx, 'Drifting', new_pos, loc[temp_idx, 1:])
                    if spos - cut_out[0] >= 0 and spos + cut_out[1] <= n_samples:
                        recordings[:, spos - cut_out[0]:spos + cut_out[1]] += amp_mod[pos] \
                                                                                                  * temp_jitt
                    elif spos - cut_out[0] < 0:
                        diff = -(spos - cut_out[0])
                        recordings[:, :spos + cut_out[1]] += amp_mod[pos] * temp_jitt[:, diff:]
                    else:
                        diff = n_samples - (spos - cut_out[0])
                        recordings[:, spos - cut_out[0]:] += amp_mod[pos] * temp_jitt[:, :diff]
            else:
                print('Electrode modulation')
                for pos, spos in enumerate(spike_pos):
                    sp_time = spos / fs
                    if sp_time < t_start_drift:
                        temp_idx = 0
                        temp_jitt = template[temp_idx, rand_idx]
                        print(sp_time, 'No drift', loc[0])
                        if spos - cut_out[0] >= 0 and spos + cut_out[1] <= n_samples:
                            recordings[:, spos - cut_out[0]:spos + cut_out[1]] += \
                                [a * t for (a, t) in zip(amp_mod[pos], temp_jitt)]
                        elif spos - cut_out[0] < 0:
                            diff = -(spos - cut_out[0])
                            recordings[:, :spos + cut_out[1]] += \
                                [a * t for (a, t) in zip(amp_mod[pos], temp_jitt[:, diff:])]
                            # recordings[:, :spos + cut_out[1]] += amp_mod[pos] * template[:, diff:]
                        else:
                            diff = n_samples - (spos - cut_out[0])
                            recordings[:, spos - cut_out[0]:] += \
                                [a * t for (a, t) in zip(amp_mod[pos], temp_jitt[:, :diff])]
                    else:
                        # compute current position
                        new_pos = np.array(loc[0, 1:] + v_drift * (sp_time - t_start_drift).rescale('s').magnitude)
                        temp_idx = np.argmin([np.linalg.norm(p - new_pos) for p in loc[:, 1:]])
                        new_temp_jitt = template[temp_idx, rand_idx]
                        print(sp_time, temp_idx, 'Drifting', new_pos, loc[temp_idx, 1:])
                        if spos - cut_out[0] >= 0 and spos + cut_out[1] <= n_samples:
                            recordings[:, spos - cut_out[0]:spos + cut_out[1]] += \
                                [a * t for (a, t) in zip(amp_mod[pos], new_temp_jitt)]
                        elif spos - cut_out[0] < 0:
                            diff = -(spos - cut_out[0])
                            recordings[:, :spos + cut_out[1]] += \
                                [a * t for (a, t) in zip(amp_mod[pos], new_temp_jitt[:, diff:])]
                            # recordings[:, :spos + cut_out[1]] += amp_mod[pos] * template[:, diff:]
                        else:
                            diff = n_samples - (spos - cut_out[0])
                            recordings[:, spos - cut_out[0]:] += \
                                [a * t for (a, t) in zip(amp_mod[pos], new_temp_jitt[:, :diff])]
                for i, t in enumerate(t_steps):
                    if t < t_start_drift:
                        temp_idx = 0
                        temp_jitt = template[temp_idx, rand_idx]
                    else:
                        # compute current position
                        new_pos = np.array(loc[0, 1:] + v_drift * (t - t_start_drift.rescale('s').magnitude))
                        temp_idx = np.argmin([np.linalg.norm(p - new_pos) for p in loc[:, 1:]])
                        temp_jitt = template[temp_idx, rand_idx]

                    feat = get_EAP_features(np.squeeze(temp_jitt), ['Na'], dt=dt)
                    mixing[i] = -np.squeeze(feat['na'])
        else:
            print('No jitter')
            if not isinstance(amp_mod[0], (list, tuple, np.ndarray)):
                print('Template modulation')
                for pos, spos in enumerate(spike_pos):
                    sp_time = spos / fs
                    if sp_time < t_start_drift:
                        temp_idx = 0
                        temp = template[temp_idx]
                    else:
                        # compute current position
                        new_pos = np.array(pos[0, 1:] + v * (sp_time - t_start_drift))
                        temp_idx = np.argmin([np.linalg.norm(p - new_pos) for p in loc[:, 1:]])
                        temp = template[temp_idx]
                    if spos - cut_out[0] >= 0 and spos + cut_out[1] <= n_samples:
                        recordings[:, spos - cut_out[0]:spos + cut_out[1]] += amp_mod[pos] * temp
                    elif spos - cut_out[0] < 0:
                        diff = -(spos - cut_out[0])
                        recordings[:, :spos + cut_out[1]] += amp_mod[pos] * temp[:, diff:]
                    else:
                        diff = n_samples - (spos - cut_out[0])
                        recordings[:, spos - cut_out[0]:] += amp_mod[pos] * temp[:, :diff]

            else:
                print('Electrode modulation')
                for pos, spos in enumerate(spike_pos):
                    sp_time = spos / fs
                    if sp_time < t_start_drift:
                        temp_idx = 0
                        temp = template[temp_idx]
                    else:
                        # compute current position
                        new_pos = np.array(pos[0, 1:] + v * (sp_time - t_start_drift))
                        temp_idx = np.argmin([np.linalg.norm(p - new_pos) for p in loc[:, 1:]])
                        temp = template[temp_idx]
                    if spos - cut_out[0] >= 0 and spos + cut_out[1] <= n_samples:
                        recordings[:, spos - cut_out[0]:spos + cut_out[1]] += \
                            [a * t for (a, t) in zip(amp_mod[pos], temp)]
                    elif spos - cut_out[0] < 0:
                        diff = -(spos - cut_out[0])
                        recordings[:, :spos + cut_out[1]] += \
                            [a * t for (a, t) in zip(amp_mod[pos], temp[:, diff:])]
                        # recordings[:, :spos + cut_out[1]] += amp_mod[pos] * template[:, diff:]
                    else:
                        diff = n_samples - (spos - cut_out[0])
                        recordings[:, spos - cut_out[0]:] += \
                            [a * t for (a, t) in zip(amp_mod[pos], temp[:, :diff])]

            for i, t in enumerate(t_steps):
                if t < t_start_drift:
                    temp_idx = 0
                    temp_jitt = template[temp_idx]
                else:
                    # compute current position
                    new_pos = np.array(loc[0, 1:] + v_drift * (t - t_start_drift.rescale('s').magnitude))
                    temp_idx = np.argmin([np.linalg.norm(p - new_pos) for p in loc[:, 1:]])
                    temp_jitt = template[temp_idx]

    final_pos = loc[temp_idx]
    print('DONE: convolution with spike ', spike_id)

    return recordings, final_pos, mixing

### RECORDING OPERATION ###

def extract_wf(spiketrains, recordings, times, fs, n_pad=2):
    '''

    Parameters
    ----------
    spiketrains
    recordings
    times
    fs
    n_pad

    Returns
    -------

    '''
    n_pad = int(n_pad * pq.ms * fs.rescale('kHz'))
    unit = times[0].rescale('ms').units

    nChs, nPts = recordings.shape

    for st in spiketrains:
        sp_rec_wf = []
        sp_amp = []
        first_spike = True

        for t in st:
            idx = np.where(times > t)[0][0]
            # find single waveforms crossing thresholds
            if idx - n_pad > 0 and idx + n_pad < nPts:
                t_spike = times[idx - n_pad:idx + n_pad]
                spike_rec = recordings[:, idx - n_pad:idx + n_pad]
            elif idx - n_pad < 0:
                t_spike = times[:idx + n_pad]
                t_spike = np.pad(t_spike, (np.abs(idx - n_pad), 0), 'constant') * unit
                spike_rec = recordings[:, :idx + n_pad]
                spike_rec = np.pad(spike_rec, ((0, 0), (np.abs(idx - n_pad), 0)), 'constant')
            elif idx + n_pad > nPts:
                t_spike = times[idx - n_pad:]
                t_spike = np.pad(t_spike, (0, idx + n_pad - nPts), 'constant') * unit
                spike_rec = recordings[:, idx - n_pad:]
                spike_rec = np.pad(spike_rec, ((0, 0), (0, idx + n_pad - nPts)), 'constant')
            if first_spike:
                nsamples = len(spike_rec)
                first_spike = False

            min_amp = np.min(spike_rec)
            sp_rec_wf.append(spike_rec)
            sp_amp.append(min_amp)

        st.waveforms = np.array(sp_rec_wf)


def filter_analog_signals(anas, freq, fs, filter_type='bandpass', order=3, copy_signal=False):
    """Filters analog signals with zero-phase Butterworth filter.
    The function raises an Exception if the required filter is not stable.

    Parameters
    ----------
    anas : np.array
           2d array of analog signals
    freq : list or float
           cutoff frequency-ies in Hz
    fs : float
         sampling frequency
    filter_type : string
                  'lowpass', 'highpass', 'bandpass', 'bandstop'
    order : int
            filter order

    Returns
    -------
    anas_filt : filtered signals
    """
    from scipy.signal import butter, filtfilt
    fn = fs / 2.
    fn = fn.rescale(pq.Hz)
    freq = freq.rescale(pq.Hz)
    band = freq / fn

    b, a = butter(order, band, btype=filter_type)

    if np.all(np.abs(np.roots(a)) < 1) and np.all(np.abs(np.roots(a)) < 1):
        print('Filtering signals with ', filter_type, ' filter at ', freq, '...')
        if len(anas.shape) == 2:
            anas_filt = filtfilt(b, a, anas, axis=1)
        elif len(anas.shape) == 1:
            anas_filt = filtfilt(b, a, anas)
        return anas_filt
    else:
        raise ValueError('Filter is not stable')


### PLOTTING ###

def raster_plots(st, bintype=False, ax=None, overlap=False, labels=False, color_st=None, color=None, fs=10,
                 marker='|', mew=2, markersize=5):
    '''

    Parameters
    ----------
    st
    bintype
    ax

    Returns
    -------

    '''
    import matplotlib.pylab as plt
    if not ax:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    for i, spiketrain in enumerate(st):
        t = spiketrain.rescale(pq.s)
        if bintype:
            if spiketrain.annotations['bintype'] == 'E':
                ax.plot(t, i * np.ones_like(t), 'b', marker=marker, mew=mew, markersize=markersize, ls='')
            elif spiketrain.annotations['bintype'] == 'I':
                ax.plot(t, i * np.ones_like(t), 'r', marker=marker, mew=mew, markersize=markersize, ls='')
        else:
            if not overlap and not labels:
                if np.any(color_st):
                    import seaborn as sns
                    colors = sns.color_palette("Paired", len(color_st))
                    if i in color_st:
                        idx = np.where(color_st==i)[0][0]
                        ax.plot(t, i * np.ones_like(t), marker=marker, mew=mew, color=colors[idx], markersize=5, ls='')
                    else:
                        ax.plot(t, i * np.ones_like(t), 'k', marker=marker, mew=mew, markersize=markersize, ls='')
                elif color is not None:
                    if isinstance(color, list) or isinstance(color, np.ndarray):
                        ax.plot(t, i * np.ones_like(t), color=color[i], marker=marker, mew=mew, markersize=markersize, ls='')
                    else:
                        ax.plot(t, i * np.ones_like(t), color=color, marker=marker, mew=mew, markersize=markersize, ls='')
                else:
                    ax.plot(t, i * np.ones_like(t), 'k', marker=marker, mew=mew, markersize=markersize, ls='')
            elif overlap:
                for j, t_sp in enumerate(spiketrain):
                    if spiketrain.annotations['overlap'][j] == 'SO':
                        ax.plot(t_sp, i, 'r', marker=marker, mew=mew, markersize=markersize, ls='')
                    elif spiketrain.annotations['overlap'][j] == 'O':
                        ax.plot(t_sp, i, 'g', marker=marker, mew=mew, markersize=markersize, ls='')
                    elif spiketrain.annotations['overlap'][j] == 'NO':
                        ax.plot(t_sp, i, 'k', marker=marker, mew=mew, markersize=markersize, ls='')
            elif labels:
                for j, t_sp in enumerate(spiketrain):
                    if 'TP' in spiketrain.annotations['labels'][j]:
                        ax.plot(t_sp, i, 'g', marker=marker, mew=mew, markersize=markersize, ls='')
                    elif 'CL' in spiketrain.annotations['labels'][j]:
                        ax.plot(t_sp, i, 'y', marker=marker, mew=mew, markersize=markersize, ls='')
                    elif 'FN' in spiketrain.annotations['labels'][j]:
                        ax.plot(t_sp, i, 'r', marker=marker, mew=mew, markersize=markersize, ls='')
                    elif 'FP' in spiketrain.annotations['labels'][j]:
                        ax.plot(t_sp, i, 'm', marker=marker, mew=mew, markersize=markersize, ls='')
                    else:
                        ax.plot(t_sp, i, 'k', marker=marker, mew=mew, markersize=markersize, ls='')

    ax.axis('tight')
    ax.set_xlim([st[0].t_start.rescale(pq.s), st[0].t_stop.rescale(pq.s)])
    ax.set_xlabel('Time (ms)', fontsize=fs)
    ax.set_ylabel('Spike Train Index', fontsize=fs)
    plt.gca().tick_params(axis='both', which='major')

    return ax


def plot_templates(templates, mea, single_figure=True):
    '''

    Parameters
    ----------
    templates
    mea_pos
    mea_pitch

    Returns
    -------

    '''
    import matplotlib.pylab as plt

    n_sources = len(templates)
    fig_t = plt.figure()

    if single_figure:
        colors = plt.rcParams['axes.color_cycle']
        ax_t = fig_t.add_subplot(111)

        for n, t in enumerate(templates):
            print('Plotting spike ', n, ' out of ', n_sources)
            if len(t.shape) == 3:
                MEA.plot_mea_recording(t.mean(axis=0), mea, colors=colors[np.mod(n, len(colors))], ax=ax_t, lw=2)
            else:
                MEA.plot_mea_recording(t, mea, colors=colors[np.mod(n, len(colors))], ax=ax_t, lw=2)

    else:
        cols = int(np.ceil(np.sqrt(n_sources)))
        rows = int(np.ceil(n_sources / float(cols)))

        for n in range(n_sources):
            ax_t = fig_t.add_subplot(rows, cols, n+1)
            MEA.plot_mea_recording(templates[n], mea, ax=ax_t)

    return fig_t


def plot_waveforms(spiketrains, mea):
    '''

    Parameters
    ----------
    spiketrains
    mea

    Returns
    -------

    '''
    import matplotlib.pylab as plt

    wf = [s.waveforms for s in spiketrains]
    fig_w = plt.figure()
    ax_w = fig_w.add_subplot(111)
    colors = plt.rcParams['axes.color_cycle']

    for n, w in enumerate(wf):
        print('Plotting spike ', n, ' out of ', len(wf))
        MEA.plot_mea_recording(w.mean(axis=0), mea, colors=colors[np.mod(n, len(colors))], ax=ax_w, lw=2)

    return fig_w